
from __future__ import print_function

import os
import shutil
import abc

import muteria.common.mix as common_mix
ERROR_HANDLER = common_mix.ErrorHandler

class BaseCallbackObject(abc.ABC):
    def __init__(self, op_retval=None, pre_callback_args=None, \
                                post_callback_args=None, \
                                repository_rootdir=None, \
                                repo_executables_relpaths=None, \
                                source_files_to_objects=None, \
                                dev_tests_list=None):
        self.op_retval = op_retval
        self.pre_callback_args = pre_callback_args
        self.post_callback_args = post_callback_args
        self.repository_rootdir = repository_rootdir
        self.repo_executables_relpaths = repo_executables_relpaths
        self.source_files_to_objects = source_files_to_objects
        self.dev_tests_list = dev_tests_list
    #~ def __init__()

    def _copy_from_repo(self, file_src_dest_map, skip_none_dest=False):
        for src, dest in list(file_src_dest_map.items()):
            if dest is None:
                if skip_none_dest: 
                    continue
                ERROR_HANDLER.error_exit("dest is None while not skipped", \
                                                                    __file__)
            abs_src = os.path.join(self.repository_rootdir, src)
            if os.path.abspath(abs_src) == os.path.abspath(dest):
                ERROR_HANDLER.error_exit("src and dest are same (from repo)", \
                                                                    __file__)
            try:
                shutil.copy2(abs_src, dest)
            except PermissionError:
                os.remove(dest)
                shutil.copy2(abs_src, dest)
    #~ def _copy_from_repo()

    def _copy_to_repo(self, file_src_dest_map, skip_none_dest=False):
        for src, dest in list(file_src_dest_map.items()):
            if dest is None:
                if skip_none_dest: 
                    continue
                ERROR_HANDLER.error_exit("dest is None while not skipped", \
                                                                    __file__)
            abs_src = os.path.join(self.repository_rootdir, src)
            if os.path.abspath(abs_src) == os.path.abspath(dest):
                ERROR_HANDLER.error_exit("src and dest are same (to repo)", \
                                                                    __file__)

            abs_src_stat = None
            if os.path.isfile(abs_src):
                abs_src_stat = os.stat(abs_src)

            try:
                shutil.copy2(dest, abs_src)
            except PermissionError:
                os.remove(abs_src)
                shutil.copy2(dest, abs_src)

            if abs_src_stat is not None:
                os.utime(abs_src, \
                              (abs_src_stat.st_atime, abs_src_stat.st_mtime))
    #~ def _copy_from_repo()

    def set_op_retval(self, op_retval):
        self.op_retval = op_retval
    #~ def set_op_retval()
    
    def set_repository_rootdir(self, repository_rootdir):
        self.repository_rootdir = repository_rootdir
    #~ def set_repository_rootdir()

    def set_repo_executables_relpaths(self, repo_executables_relpaths):
        self.repo_executables_relpaths = repo_executables_relpaths
    #~ def set_repo_executables_relpaths()

    def set_source_files_to_objects(self, source_files_to_objects):
        self.source_files_to_objects = source_files_to_objects
    #~ def set_source_files_to_objects()

    def set_dev_tests_list(self, dev_tests_list):
        self.dev_tests_list = dev_tests_list
    #~ def set_dev_tests_list()

    # Following methods are called by the caller. 
    # The above are called by repo manager
    
    def set_pre_callback_args (self, pre_callback_args ):
        self.pre_callback_args = pre_callback_args 
    #~ def set_pre_callback_args()

    def set_post_callback_args (self, post_callback_args ):
        self.post_callback_args = post_callback_args 
    #~ def set_post_callback_args()

    @abc.abstractmethod
    def before_command(self):
        """ This method is executed BEFORE the execution of the
            corrresponding command.
            :return: (bool) True on success and False on failure
        """
        pass

    @abc.abstractmethod
    def after_command(self):
        """ This method is executed AFTER the execution of the
            corrresponding command. The execution of the command return 
            the boolean value that is then set in 'self.op_retval' 
            (this is set to None prior execution of the command).
            :return: (bool) True on success, False on failure.
        """
        pass
#~ class BaseCallbackObject

class DefaultCallbackObject(BaseCallbackObject):
    """ Use this class object if no call back is needed
    """
    def before_command(self):
        """ Return True for success
        """
        return common_mix.GlobalConstants.COMMAND_SUCCESS
    #~ def before_command()

    def after_command(self):
        """ Return True for success
        """
        return common_mix.GlobalConstants.COMMAND_SUCCESS
    #~ def after_command()
#~ class DefaultCallbackObject




from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import resource

import muteria.common.fs as common_fs
import muteria.common.mix as common_mix

from muteria.repositoryandcode.codes_convert_support import CodeFormats
from muteria.drivers.testgeneration.base_testcasetool import BaseTestcaseTool
from muteria.drivers.testgeneration.testcases_info import TestcasesInfoObject
from muteria.drivers import DriversUtils
from muteria.drivers.testgeneration.testcase_formats.ktest.ktest \
                                                        import KTestTestFormat

ERROR_HANDLER = common_mix.ErrorHandler

class TestcasesToolKlee(BaseTestcaseTool):

    @classmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        for prog in ('klee',):
            if custom_binary_dir is not None:
                prog = os.path.join(custom_binary_dir, prog)
            if not DriversUtils.check_tool(prog=prog, args_list=['--version'],\
                                                    expected_exit_codes=[0,1]):
                return False

        return KTestTestFormat.installed(custom_binary_dir=custom_binary_dir)
    #~ def installed()

    def __init__(self, *args, **kwargs):
        BaseTestcaseTool.__init__(self, *args, **kwargs)
        self.test_details_file = \
                    os.path.join(self.tests_working_dir, 'test_details.json')
        self.klee_used_tmp_build_dir = os.path.join(self.tests_working_dir, \
                                    self._get_tool_name()+'_used_tmp_build_dir')

        # mapping between exes, to have a local copy for execution
        self.repo_exe_to_local_to_remote = {}

        self.keptktest2dupktests = os.path.join(self.tests_working_dir, \
                                                'kept_to_dup_ktests_map.json')

        if os.path.isdir(self.klee_used_tmp_build_dir):
            shutil.rmtree(self.klee_used_tmp_build_dir)
        os.mkdir(self.klee_used_tmp_build_dir)
    #~ def __init__()

    # SHADOW override
    def _get_default_params(self):
        bool_params = {
            '-ignore-solver-failures': None,
            '-allow-external-sym-calls': True, #None,
            '-posix-runtime': True, #None,
            '-dump-states-on-halt': True, #None,
            '-only-output-states-covering-new': True,
            '-use-cex-cache': True,
        }
        key_val_params = {
            '-output-dir': self.tests_storage_dir,
            '-solver-backend': None,
            '-max-solver-time': None,
            '-search': None,
            '-max-memory': None,
            '-max-time': self.config.TEST_GENERATION_MAXTIME,
            '-libc': 'uclibc',
            '-max-sym-array-size': '4096',
            '-max-instruction-time': '10.',
        }
        return bool_params, key_val_params
    #~ def _get_default_params()
    
    # SHADOW override
    def _get_sym_args(self):
        # sym args
        default_sym_args = ['-sym-arg', '5']

        klee_sym_args = default_sym_args
        uc = self.config.get_tool_user_custom()
        if uc is not None:
            post_bc_cmd = uc.POST_TARGET_CMD_ORDERED_FLAGS_LIST
            if post_bc_cmd is not None:
                klee_sym_args = []
                for tup in post_bc_cmd:
                    klee_sym_args += list(tup)
        return klee_sym_args
    #~ def _get_sym_args()

    # SHADOW should override
    def _get_back_llvm_compiler(self):
        return None #'clang'
    #~ def _get_back_llvm_compiler()

    # SHADOW should override
    def _get_back_llvm_compiler_path(self):
        return None 
    #~ def _get_back_llvm_compiler_path()

    @staticmethod
    def get_value_in_arglist(arglist, flagname):
        value = None
        for i, v in enumerate(arglist):
            if v in ('-'+flagname, '--'+flagname):
                value = arglist[i+1]
                break
            elif v.startswith('-'+flagname+'=') \
                                            or v.startswith('--'+flagname+'='):
                _, value = v.split('=')
                break
        return value
    #~ def get_value_in_arglist()

    @staticmethod
    def set_value_in_arglist(arglist, flagname, value):
        for i, v in enumerate(arglist):
            if v in ('-'+flagname, '--'+flagname):
                arglist[i+1] = value
                break
            elif v.startswith('-'+flagname+'=') \
                                            or v.startswith('--'+flagname+'='):
                pre, _ = v.split('=')
                arglist[i] = pre + '=' + str(value)
                break
    #~ def set_value_in_arglist()

    @staticmethod
    def remove_arg_and_value_from_arglist(arglist, flagname):
        for i, v in enumerate(arglist):
            if v in ('-'+flagname, '--'+flagname):
                del arglist[i+1]
                del arglist[i]
                break
            elif v.startswith('-'+flagname+'=') \
                                            or v.startswith('--'+flagname+'='):
                del arglist[i]
                break
    #~ def remove_arg_and_value_from_arglist()

    # SHADOW should override
    def _call_generation_run(self, runtool, args):
        ## locate max-time
        timeout_grace_period = 600
        max_time = None
        cur_max_time = float(self.get_value_in_arglist(args, 'max-time'))
        max_time = cur_max_time + 5 #5 to give time to klee

        # set stack to unlimited
        stack_ulimit_soft, stack_ulimit_hard = \
                                    resource.getrlimit(resource.RLIMIT_STACK)
        if stack_ulimit_soft != -1:
            resource.setrlimit(resource.RLIMIT_STACK, (-1, stack_ulimit_hard))

        # Execute Klee
        ret, out, err = DriversUtils.execute_and_get_retcode_out_err(\
                                    runtool, args, timeout=max_time, \
                                    timeout_grace_period=timeout_grace_period)
                                    #out_on=False, err_on=False)
        
        # restore stack
        if stack_ulimit_soft != -1:
            resource.setrlimit(resource.RLIMIT_STACK, \
                                        (stack_ulimit_soft, stack_ulimit_hard))

        if (ret != 0 and ret not in DriversUtils.EXEC_TIMED_OUT_RET_CODE):
            logging.error(out)
            logging.error(err)
            logging.error("\n>> CMD: " + " ".join([runtool]+args) + '\n')
            ERROR_HANDLER.error_exit("call to klee testgen failed, "
                                    + "error code is {}".format(ret), __file__)
    #~ def _call_generation_run()

    # SHADOW should override
    def _get_testexec_extra_env_vars(self, testcase):
        return None
    #~ def _get_testexec_extra_env_vars()
    
    def _get_tool_name(self):
        return 'klee'
    #~ def _get_tool_name()

    def _get_input_bitcode_file(self, code_builds_factory, rel_path_map, \
                                                meta_criteria_tool_obj=None):
        back_llvm_compiler = self._get_back_llvm_compiler() 
        back_llvm_compiler_path = self._get_back_llvm_compiler_path() 
        
        pre_ret, ret, post_ret = code_builds_factory.transform_src_into_dest(\
                        src_fmt=CodeFormats.C_SOURCE,\
                        dest_fmt=CodeFormats.LLVM_BITCODE,\
                        src_dest_files_paths_map=rel_path_map,\
                        compiler=back_llvm_compiler, \
                        llvm_compiler_path=back_llvm_compiler_path, \
                        clean_tmp=True, reconfigure=True)
        if ret == common_mix.GlobalConstants.COMMAND_FAILURE:
            ERROR_HANDLER.error_exit("Program {}.".format(\
                                        'LLVM built problematic'), __file__)

        # Update exe_map to reflect bitcode extension
        rel2bitcode = {}
        for r_file, b_file in list(rel_path_map.items()):
            bc = b_file+'.bc'
            ERROR_HANDLER.assert_true(os.path.isfile(bc), \
                                    "Bitcode file not existing: "+bc, __file__)
            rel2bitcode[r_file] = bc

        ERROR_HANDLER.assert_true(len(rel_path_map) == 1, \
                            "Support single bitcode module for now", __file__)

        bitcode_file = rel2bitcode[list(rel2bitcode.keys())[0]]
        return bitcode_file
    #~ def _get_input_bitcode_file()
    ########################################################################

    def get_testcase_info_object(self):
        try:
            return self.testcase_info_object
        except AttributeError:
            tc_info_obj = TestcasesInfoObject()
            cwd = os.getcwd()
            os.chdir(self.tests_storage_dir)
            for root, _, files in os.walk('.'):
                for f in files:
                    tc = os.path.normpath(os.path.join(root, f))
                    if tc.endswith(KTestTestFormat.ktest_extension):
                        tc_info_obj.add_test(tc)
            os.chdir(cwd)
            self.testcase_info_object = tc_info_obj
            return self.testcase_info_object
    #~ def get_testcase_info_object()

    def _prepare_executable(self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Make sure we have the right executable ready (if needed)
        """
        #self.code_builds_factory.copy_into_repository(exe_path_map)
        pass
    #~ def _prepare_executable()

    def _restore_default_executable(self, exe_path_map, env_vars, \
                                                        collect_output=False):
        """ Restore back the default executable (if needed).
            Useful for test execution that require the executable
            at a specific location.
        """
        #self.code_builds_factory.restore_repository_files(exe_path_map)
        pass
    #~ def _restore_default_executable()

    def _execute_a_test (self, testcase, exe_path_map, env_vars, \
                    callback_object=None, timeout=None, collect_output=False):
        """ Execute a test given that the executables have been set 
            properly
        """

        if timeout is None:
            timeout = self.config.ONE_TEST_EXECUTION_TIMEOUT
        
        #logging.debug('TIMEOUT: '+str(timeout))

        ERROR_HANDLER.assert_true(len(exe_path_map) == 1, \
                                    "support a single exe for now", __file__)
        ERROR_HANDLER.assert_true(callback_object is None, \
                                        'TODO: handle callback_obj', __file__)
        
        repo_exe = list(exe_path_map.keys())[0]
        local_exe = os.path.join(self.klee_used_tmp_build_dir, repo_exe)
        if repo_exe not in self.repo_exe_to_local_to_remote:
            with self.shared_loc:
                if repo_exe not in self.repo_exe_to_local_to_remote:
                    if not os.path.isdir(os.path.dirname(local_exe)):
                        os.makedirs(os.path.dirname(local_exe))
                    self.repo_exe_to_local_to_remote[repo_exe] = \
                                                            {local_exe: None}

        remote_exe = exe_path_map[repo_exe]
        if remote_exe is None:
            remote_exe = repo_exe

        if remote_exe != self.repo_exe_to_local_to_remote[repo_exe][local_exe]:
            with self.shared_loc:
                if remote_exe != \
                        self.repo_exe_to_local_to_remote[repo_exe][local_exe]:
                    if remote_exe == repo_exe:
                        self.code_builds_factory.set_repo_to_build_default(\
                                        also_copy_to_map={repo_exe: local_exe})
                    else:
                        shutil.copy2(remote_exe, local_exe)
                        # Use hard link to avoid copying for big files
                        #if os.path.isfile(local_exe):
                        #    os.remove(local_exe)
                        #os.link(remote_exe, local_exe)
                    self.repo_exe_to_local_to_remote[repo_exe][local_exe] = \
                                                                    remote_exe

        collected_output = [] if collect_output else None

        extra_env = self._get_testexec_extra_env_vars(testcase)
        if extra_env is not None and len(extra_env) > 0:
            env_vars = dict(env_vars).update(extra_env)

        verdict = KTestTestFormat.execute_test(local_exe, \
                        os.path.join(self.tests_storage_dir, testcase), \
                        env_vars=env_vars, \
                        timeout=timeout, \
                        collected_output=collected_output, \
                        custom_replay_tool_binary_dir=self.custom_binary_dir)

        return verdict, collected_output
    #~ def _execute_a_test()

    def _do_generate_tests (self, exe_path_map, code_builds_factory, \
                                                meta_criteria_tool_obj=None, \
                                                                max_time=None):
        # Check the passed exe_path_map
        if exe_path_map is not None:
            for r_exe, v_exe in exe_path_map.items():
                ERROR_HANDLER.assert_true(v_exe is None, \
                            "Klee driver does not use passed exe_path_map", \
                                                                    __file__)

        # Setup
        if os.path.isdir(self.tests_working_dir):
            try:
                shutil.rmtree(self.tests_working_dir)
            except PermissionError:
                self._dir_chmod777(self.tests_working_dir)
                shutil.rmtree(self.tests_working_dir)

        os.mkdir(self.tests_working_dir)
        if os.path.isdir(self.tests_storage_dir):
            shutil.rmtree(self.tests_storage_dir)
        
        prog = self._get_tool_name()
        if self.custom_binary_dir is not None:
            prog = os.path.join(self.custom_binary_dir, prog)
            ERROR_HANDLER.assert_true(os.path.isfile(prog), \
                            "The tool {} is missing from the specified dir {}"\
                                        .format(os.path.basename(prog), \
                                            self.custom_binary_dir), __file__)

        rel_path_map = {}
        exes, _ = code_builds_factory.repository_manager.\
                                                    get_relative_exe_path_map()
        for exe in exes:
            filename = os.path.basename(exe)
            rel_path_map[exe] = os.path.join(self.tests_working_dir, filename)

        # Get bitcode file
        bitcode_file = self._get_input_bitcode_file(code_builds_factory, \
                                                                rel_path_map, \
                                meta_criteria_tool_obj=meta_criteria_tool_obj)
        repo_rel_exe_file = list(rel_path_map)[0]
        
        # klee params
        bool_param, k_v_params = self._get_default_params()
        if max_time is not None:
            k_v_params['-max-time'] = str(max_time)

        # Consider pre user custom
        uc = self.config.get_tool_user_custom()
        pre_args = []
        if uc is not None:
            pre_bc_cmd = uc.PRE_TARGET_CMD_ORDERED_FLAGS_LIST
            if pre_bc_cmd is not None:
                for tup in pre_bc_cmd:
                    key = tup[0][1:] if tup[0].startswith('--') else tup[0]
                    bool_disabled = False
                    if key in k_v_params:
                        del k_v_params[key]
                    if key in bool_param:
                        del bool_param[key]
                        if len(tup) == 2:
                            ERROR_HANDLER.assert_true(\
                                        type(tup[1]) == bool, \
                                        "bool arg is either enabled or "
                                    "disabled. val must be bool", __file__)
                            if not tup[1]:
                                bool_disabled = True
                            tup = tup[:1]
                        else:
                            ERROR_HANDLER.assert_true(len(tup) == 1, \
                                    "Invalid bool param: "+key, __file__)
                    if not bool_disabled:
                        pre_args.extend(list(tup))

        args = [bp for bp, en in list(bool_param.items()) if en]
        for k,v in list(k_v_params.items()):
            if v is not None:
                args += [k,str(v)]
        args.extend(pre_args)

        args.append(bitcode_file)

        args += self._get_sym_args()

        self._call_generation_run(prog, list(args))

        # XXX: remove duplicate tests
        kepttest2duptest_map = {}
        dup_list, invalid = KTestTestFormat.ktest_fdupes(\
                                                    self.tests_storage_dir, \
                        custom_replay_tool_binary_dir=self.custom_binary_dir)
        if len(invalid) > 0:
            logging.warning(\
                        "{} generated ktests are invalid".format(len(invalid)))
            for kt in invalid:
                os.remove(kt)
        for dup_tuple in dup_list:
            kepttest2duptest_map[os.path.relpath(\
                            dup_tuple[0], \
                            self.tests_storage_dir)] = [os.path.relpath(dp) \
                                                    for dp in dup_tuple[1:]]
            for df in dup_tuple[1:]:
                os.remove(df)
        common_fs.dumpJSON(kepttest2duptest_map, self.keptktest2dupktests)
        
        # Copy replay tool into test folder
        klee_replay_pathname = KTestTestFormat.get_test_replay_tool(\
                        custom_replay_tool_binary_dir=self.custom_binary_dir)
        shutil.copy2(klee_replay_pathname, self.tests_storage_dir)

        store_obj = {repo_rel_exe_file: os.path.basename(bitcode_file)}
        common_fs.dumpJSON(store_obj, self.test_details_file)
    #~ def _do_generate_tests()

    def can_run_tests_in_parallel(self):
        return True
    #~ def can_run_tests_in_parallel()
#~ class TestcasesToolKlee


#
# [LICENCE]
#
""" Code coverage tool module. The class of interest is BaseCOdecoverageTool.

The tools are organized by programming language

For each language, there is a folder for each tool, 
named after the tool in all lowercase , starting with letter or underscore(_),
The remaining caracters are either letter, number or underscore

XXX Each criteria tool package must have the 
following in the __init__.py file:
>>> import <Module>.<class extending BaseCodecoverageTool> as CodecoverageTool
"""

from __future__ import print_function
import os
import sys
import glob
import shutil
import logging
import abc
import tqdm

import muteria.common.matrices as common_matrices
import muteria.common.mix as common_mix

from muteria.drivers.checkpoint_handler import CheckPointHandler
from muteria.drivers.criteria import TestCriteria
from muteria.drivers import DriversUtils

ERROR_HANDLER = common_mix.ErrorHandler

class BaseCriteriaTool(abc.ABC):
    '''
    '''
    @classmethod
    def get_supported_criteria(cls):
        return cls._get_meta_instrumentation_criteria() + \
                                cls._get_separated_instrumentation_criteria()
    #~ def get_supported_criteria()

    def __init__(self, meta_test_generation_obj, criteria_working_dir, 
                                code_builds_factory, config, checkpointer):
        # Set Constants
        
        # Set Direct Arguments Variables
        self.meta_test_generation_obj = meta_test_generation_obj
        self.criteria_working_dir = criteria_working_dir
        self.code_builds_factory = code_builds_factory
        self.config = config
        self.checkpointer = checkpointer
        
        # Verify Direct Arguments Variables
        ERROR_HANDLER.assert_true(self.criteria_working_dir is not None, \
                            "Must specify criteria_working_dir", __file__)
        
        # Set Indirect Arguments Variables
        ## put the instrumented code into this folder (to be created by user)
        self.instrumented_code_storage_dir = os.path.join(
                        self.criteria_working_dir, "instrumented_code")
        self.custom_binary_dir = None
        if self.config.tool_user_custom is not None:
            self.custom_binary_dir = \
                        self.config.tool_user_custom.PATH_TO_TOOL_BINARY_DIR

        
        # Verify indirect Arguments Variables
        
        # Initialize Other Fields
        
        # Make Initialization Computation
        ## Create dirs
        if not os.path.isdir(self.criteria_working_dir):
            self.clear_working_dir()
    #~ def __init__()

    def clear_working_dir(self):
        if os.path.isdir(self.criteria_working_dir):
            try:
                shutil.rmtree(self.criteria_working_dir)
            except PermissionError:
                self._dir_chmod777(self.criteria_working_dir)
                shutil.rmtree(self.criteria_working_dir)
        os.mkdir(self.criteria_working_dir)
    #~ def clear_working_dir(self):

    def get_toolname(self):
        return self.config.get_tool_name()
    #~ def get_toolname()

    def get_toolalias(self):
        return self.config.get_tool_config_alias()
    #~ def get_toolalias()

    def get_checkpointer(self):
        return self.checkpointer
    #~ def get_checkpointer()

    def has_checkpointer(self):
        return self.checkpointer is not None
    #~ def has_checkpointer()

    def _runtest_meta_criterion_program (self, testcases, criterion_to_matrix,\
                                    criterion_to_executionoutput, \
                                    criteria_element_list_by_criteria, \
                                    cover_criteria_elements_once=False, \
                                    prioritization_module_by_criteria=None, \
                                    test_parallel_count=1):
        """
        """

        logging.debug("# Executing meta {}: {} ...".format("criteria" if \
                                len(criterion_to_matrix) > 1 else "criterion",\
                                [c.get_str() for c in criterion_to_matrix]))

        # get instrumented executables
        criterion2executable_path = \
                                self.get_instrumented_executable_paths_map( \
                                                    criterion_to_matrix.keys())

        # get environment vars
        result_dir_tmp = os.path.join(self.criteria_working_dir, \
                                                "criteria_meta_result_tmp")
        if os.path.isdir(result_dir_tmp):
            try:
                shutil.rmtree(result_dir_tmp)
            except PermissionError:
                self._dir_chmod777(result_dir_tmp)
                shutil.rmtree(result_dir_tmp)

        criterion2environment_vars = self._get_criteria_environment_vars( \
                                                            result_dir_tmp, \
                                enabled_criteria=criterion_to_matrix.keys())

        ERROR_HANDLER.assert_true(set(criterion2executable_path) == \
                                            set(criterion2environment_vars), \
                        "mismatch between exe_path_map and env_vars", __file__)

        criterialist = criterion2executable_path.keys()

        # group criteria
        groups = self._get_criteria_groups(criterion2executable_path,\
                                                    criterion2environment_vars)

        # important tmp vars
        criterion2coverage_per_test = \
                                {criterion: {} for criterion in criterialist}
        criterion2metaoutlog_per_test = {}
        a_criterion_has_outlog = False
        for criterion in criterialist:
            if criterion_to_executionoutput is None:
                criterion2metaoutlog_per_test[criterion] = None
            else:
                criterion2metaoutlog_per_test[criterion] = {}
                a_criterion_has_outlog = True

        timeout_times = self.config.META_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES

        # Execute each test and gather the data
        processbar = tqdm.tqdm(testcases, leave=False, dynamic_ncols=True) 
        for testcase in processbar: 
            processbar.set_description("Running Test %s"% testcase)
            for cg_criteria, cg_exe_path_map, cg_env_vars in groups:
                # Create reult_tmp_dir
                os.mkdir(result_dir_tmp, mode=0o777)

                # run testcase
                test_verdict = self.meta_test_generation_obj.execute_testcase(\
                                            testcase, \
                                            exe_path_map=cg_exe_path_map, \
                                            env_vars=cg_env_vars,\
                                            use_recorded_timeout_times=\
                                                                timeout_times)
                
                # Collect temporary data into result_dir_tmp
                self._collect_temporary_coverage_data(\
                                                cg_criteria, test_verdict, \
                                                cg_env_vars, result_dir_tmp, \
                                                testcase)

                # extract coverage
                coverage_tmp_data_per_criterion = \
                                self._extract_coverage_data_of_a_test(\
                                                cg_criteria, test_verdict, \
                                                    result_dir_tmp)
                if a_criterion_has_outlog:
                    metaoutlog_tmp_data_per_criterion = \
                                    self._extract_metaoutlog_data_of_a_test( \
                                                cg_criteria, test_verdict, \
                                                    result_dir_tmp)
                # update data
                for criterion in cg_criteria:
                    if len(criterion2coverage_per_test[criterion]) == 0:
                        for elem in coverage_tmp_data_per_criterion[criterion]:
                            criterion2coverage_per_test[criterion][elem] = {}
                        if criterion_to_executionoutput[criterion] is not None:
                            for elem in metaoutlog_tmp_data_per_criterion[\
                                                                    criterion]:
                                criterion2metaoutlog_per_test[criterion]\
                                                                    [elem] = {}

                    for elem in coverage_tmp_data_per_criterion[criterion]:
                        # verify that the value is positive or null
                        v_elem = coverage_tmp_data_per_criterion[criterion]\
                                                                        [elem]
                        ERROR_HANDLER.assert_true(type(v_elem) == int, \
                                        "cov num type must be int", __file__)
                        ERROR_HANDLER.assert_true(v_elem >= 0, \
                                        "invalid cov num(negative)", __file__)
                        try:
                            res = criterion2coverage_per_test[criterion][elem]
                        except KeyError:
                            res = {}
                            criterion2coverage_per_test[criterion][elem] = res

                        res[testcase] = coverage_tmp_data_per_criterion\
                                                            [criterion][elem]

                    if criterion_to_executionoutput[criterion] is not None:
                        for elem in metaoutlog_tmp_data_per_criterion[\
                                                                    criterion]:
                            try:
                                res = criterion2metaoutlog_per_test[criterion]\
                                                                        [elem]
                            except KeyError:
                                res = {}
                                criterion2metaoutlog_per_test[criterion][\
                                                                    elem] = res

                            res[testcase] = metaoutlog_tmp_data_per_criterion\
                                                            [criterion][elem]
                            

                # remove dir created for temporal storage
                try:
                    shutil.rmtree(result_dir_tmp)
                except PermissionError:
                    self._dir_chmod777(result_dir_tmp)
                    shutil.rmtree(result_dir_tmp)

        # Write the execution data into the matrices
        # Since for ExecutionMatrix, active is not 0 thus this is direct.
        testcases_set = set(testcases)
        for criterion in criterion2coverage_per_test:
            matrix = criterion_to_matrix[criterion]
            matrix_file = matrix.get_store_filename()
            if matrix_file is not None and os.path.isfile(matrix_file):
                os.remove(matrix_file)
                                    
            for key, value in list(\
                            criterion2coverage_per_test[criterion].items()):
                missing_tests = {t:common_mix.GlobalConstants.\
                                        ELEMENT_NOTCOVERED_VERDICT for t in \
                                                    testcases_set - set(value)}
                value.update(missing_tests)
                matrix.add_row_by_key(key, value, serialize=False)
            # Serialize to disk
            matrix.serialize()

            # outLog
            if criterion_to_executionoutput[criterion] is not None:
                outlog_datfile = criterion_to_executionoutput[criterion]\
                                                        .get_store_filename()
                if outlog_datfile is not None and \
                                                os.path.isfile(outlog_datfile):
                    os.remove(outlog_datfile)

                criterion_to_executionoutput[criterion].add_data(\
                                    criterion2metaoutlog_per_test[criterion], \
                                    serialize=True)
    #~ def _runtest_meta_criterion_program()

    def _runtest_separate_criterion_program (self, criterion, testcases, \
                                    matrix, 
                                    executionoutput, \
                                    criteria_element_list, \
                                    cover_criteria_elements_once=False, \
                                    prioritization_module=None, \
                                    test_parallel_count=1,
                                    serialize_period=5,
                                    checkpoint_handler=None, \
                                    cp_calling_func_name=None, \
                                    cp_calling_done_task_id=None, \
                                    cp_calling_tool=None):
        '''
        Note: Here the temporary matrix is used as checkpoint 
                (with frequency the 'serialize_period' parameter). 
            The checkpointer is mainly used for the execution time
            (TODO: support parallelism: per test outdata)
        '''
        # FIXME: Support parallelism, then remove the code
        # bellow:
        ERROR_HANDLER.assert_true(test_parallel_count <= 1, \
                    "FIXME: Must first implement support for parallel mutatio")
        #~ FXIMEnd

        # @Checkpoint: validate
        if checkpoint_handler is not None:
            ERROR_HANDLER.assert_true(cp_calling_func_name is not None)
            ERROR_HANDLER.assert_true(cp_calling_done_task_id is not None)
            ERROR_HANDLER.assert_true(cp_calling_tool is not None)
            cp_data = checkpoint_handler.get_optional_payload()
            if cp_data is None:
                cp_data = [{}, {}]
        else:
            cp_data = [{}, {}]

        failverdict_to_val_map = {
                    common_mix.GlobalConstants.FAIL_TEST_VERDICT: \
                                            matrix.getActiveCellDefaultVal(),
                    common_mix.GlobalConstants.PASS_TEST_VERDICT: \
                                            matrix.getInactiveCellVal(), 
                    common_mix.GlobalConstants.UNCERTAIN_TEST_VERDICT: \
                                            matrix.getUncertainCellDefaultVal()
        }

        assert serialize_period >= 1, \
                            "Serialize period must be an integer in [1,inf["

        # matrix based checkpoint
        completed_elems = set(cp_data[0])

        if criteria_element_list is None:
            criteria_element_list = self.get_criterion_info_object(criterion).\
                                                            get_elements_list()

        ERROR_HANDLER.assert_true(prioritization_module is not None, 
                                        "prioritization module must be passed")
            
        timeout_times = \
                    self.config.SEPARATED_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES

        # in case the test list is empty, do nothing
        if len(testcases) > 0:
            # main loop for elements execution
            num_elems = len(criteria_element_list)
            pos = -1 + len(completed_elems)
            ## prepare the optimizer
            prioritization_module.reset(self.config.get_tool_config_alias(), \
                                            criteria_element_list, testcases)
            while prioritization_module.has_next_test_objective():
                #pos += 1 # Done bellow in the case where it was not completed
                element = prioritization_module.get_next_test_objective()

                # @Checkpointing: check if already executed
                if element in completed_elems:
                    continue
                else:
                    pos += 1

                logging.debug("# Executing {} element {} ({}/{}) ...".format( \
                                    criterion.get_str(), \
                                    DriversUtils.make_meta_element(element, \
                                        self.config.get_tool_config_alias()), \
                                    pos, num_elems))

                # execute element with the given testcases
                element_executable_path = \
                                self._get_criterion_element_executable_path(\
                                                            criterion, element)
                execution_environment_vars = \
                                self._get_criterion_element_environment_vars(\
                                                            criterion, element)
                # run optimizer with all tests of targeting the test objective
                may_cov_tests = prioritization_module\
                                        .get_test_execution_optimizer(element)\
                                        .select_tests(100, is_proportion=True)

                cannot_cov_tests = set(testcases) - set(may_cov_tests)
                
                fail_verdicts, exec_outs_by_tests = \
                                    self.meta_test_generation_obj.runtests(\
                                        meta_testcases=may_cov_tests, \
                                        exe_path_map=element_executable_path, \
                                        env_vars=execution_environment_vars, \
                                        stop_on_failure=\
                                                cover_criteria_elements_once, \
                                        use_recorded_timeout_times=\
                                                            timeout_times, \
                                        with_output_summary=(executionoutput \
                                                                is not None), \
                                        parallel_test_count=None, \
                                        restart_checkpointer=True)
                prioritization_module.feedback(element, fail_verdicts)

                fail_verdicts.update({\
                            v: common_mix.GlobalConstants.PASS_TEST_VERDICT \
                                                for v in cannot_cov_tests})
                # put in row format for matrix
                matrix_row_key = element
                matrix_row_values = \
                                {tc:failverdict_to_val_map[fail_verdicts[tc]] \
                                                    for tc in fail_verdicts}
                serialize_on = (pos % serialize_period == 0)
                cp_data[0][matrix_row_key] = matrix_row_values

                if executionoutput is not None:
                    cp_data[1][element] = exec_outs_by_tests

                # @Checkpointing: for time
                if serialize_on and checkpoint_handler is not None:
                    checkpoint_handler.do_checkpoint( \
                                            func_name=cp_calling_func_name, \
                                            taskid=cp_calling_done_task_id, \
                                            tool=cp_calling_tool, \
                                            opt_payload=cp_data)

        # Write the execution data into the matrix
        for matrix_row_key, matrix_row_values in list(cp_data[0].items()):
            matrix.add_row_by_key(matrix_row_key, matrix_row_values, \
                                                            serialize=False)

        # TODO: Make the following two instructions atomic                                                    
        # final serialization (in case #Muts not multiple od serialize_period)
        matrix.serialize()

        # Write the execution output data
        if executionoutput is not None:
            if len(cp_data[1]) > 0:
                executionoutput.add_data(cp_data[1], serialize=True)
            else:
                executionoutput.serialize()
    #~ def _runtest_separate_criterion_program()

    def runtests_criteria_coverage (self, testcases, \
                                    criteria_element_list_by_criteria, \
                                    criterion_to_matrix, \
                                    criterion_to_executionoutput, \
                                    re_instrument_code=True, \
                                    cover_criteria_elements_once=False, \
                                    prioritization_module_by_criteria=None, \
                                    test_parallel_count=1):
        """
            (TODO: support parallelism: per test outdata)
        """
        # FIXME: Support parallelism, then remove the code
        # bellow:
        ERROR_HANDLER.assert_true(test_parallel_count <= 1, \
                "FIXME: Must first implement support for parallel mutation")
        #~ FXIMEnd

        # @Checkpoint: create a checkpoint handler (for time)
        cp_func_name = "runtests_criteria_coverage"
        cp_task_id = 1
        checkpoint_handler = CheckPointHandler(self.get_checkpointer())
        if checkpoint_handler.is_finished():
            return

        ERROR_HANDLER.assert_true(len(criterion_to_matrix) > 0, \
                                            "no criterion enabled", __file__)

        ERROR_HANDLER.assert_true(len(set(criterion_to_matrix) - \
                                set(self.get_supported_criteria())) == 0, \
                            "Some unsuported criteria are enabled", __file__)

        ERROR_HANDLER.assert_true(criterion_to_executionoutput is None or \
                set(criterion_to_matrix) == set(criterion_to_executionoutput),\
                "mismatch of criteria between matrix and outlog" , __file__)

        # Check that the result_matrix is empty and fine
        for criterion in criterion_to_matrix:
            ERROR_HANDLER.assert_true( \
                            criterion_to_matrix[criterion].is_empty(), \
                                          "the matrix must be empty", __file__)

            ERROR_HANDLER.assert_true( \
                    set(testcases) == set(criterion_to_matrix[criterion]\
                                                .get_nonkey_colname_list()), \
                        "The specified test cases are not same in the matrix",
                                                                    __file__)

            ERROR_HANDLER.assert_true(\
                        criterion_to_executionoutput[criterion] is None or \
                        criterion_to_executionoutput[criterion].is_empty(), \
                                    "the execoutput must be empty", __file__)

        # @Checkpoint: check 
        if checkpoint_handler.is_to_execute(func_name=cp_func_name, \
                                                            taskid=cp_task_id):
            # Intrument the codes is requested 
            if re_instrument_code:
                self.instrument_code(\
                                enabled_criteria=criterion_to_matrix.keys())

            # @Checkpoint: checkpoint
            checkpoint_handler.do_checkpoint(func_name=cp_func_name, \
                                                taskid=cp_task_id)

        # @Checkpoint: next task
        cp_task_id += 1

        # Split criteria
        meta_crits = list(set(criterion_to_matrix) & \
                                set(self._get_meta_instrumentation_criteria()))
        separated_crits = list(set(criterion_to_matrix) & \
                        set(self._get_separated_instrumentation_criteria()))

        m_crit2mat, s_crit2mat = self._extract_sub_dicts(criterion_to_matrix,\
                                                [meta_crits, separated_crits])
        if criterion_to_executionoutput is not None:
            m_crit2execout, s_crit2execout = self._extract_sub_dicts(\
                                                criterion_to_executionoutput,\
                                                [meta_crits, separated_crits])
        m_crit2elem, s_crit2elem = self._extract_sub_dicts( \
                                        criteria_element_list_by_criteria,\
                                                [meta_crits, separated_crits])
        m_crit2pm, s_crit2pm = None, None
        if prioritization_module_by_criteria is not None:
            m_crit2pm, s_crit2pm = self._extract_sub_dicts( \
                                            prioritization_module_by_criteria,\
                                                [meta_crits, separated_crits])
        else:
            s_crit2pm = {c: None for c in separated_crits}

        # runtest with the meta files
        if len(m_crit2mat) > 0 and \
                    checkpoint_handler.is_to_execute(func_name=cp_func_name, \
                                            taskid=cp_task_id):
            self._runtest_meta_criterion_program(testcases=testcases, \
                                criterion_to_matrix=m_crit2mat, \
                                criterion_to_executionoutput=m_crit2execout, \
                                criteria_element_list_by_criteria=m_crit2elem,\
                                cover_criteria_elements_once=\
                                            cover_criteria_elements_once,\
                                prioritization_module_by_criteria=m_crit2pm, \
                                test_parallel_count=test_parallel_count)

            # @Checkpoint: checkpoint
            checkpoint_handler.do_checkpoint(func_name=cp_func_name, \
                                                taskid=cp_task_id)

        # @Checkpoint: next task
        cp_task_id += 1

        # runtest with the separate files
        for criterion in s_crit2mat.keys():
            if checkpoint_handler.is_to_execute(func_name=cp_func_name, \
                                taskid=cp_task_id, tool=criterion.get_str()):
                self._runtest_separate_criterion_program(criterion, \
                                testcases=testcases, \
                                matrix=s_crit2mat[criterion], \
                                executionoutput=s_crit2execout[criterion], \
                                criteria_element_list=s_crit2elem[criterion],\
                                cover_criteria_elements_once=\
                                            cover_criteria_elements_once,\
                                prioritization_module=s_crit2pm[criterion], \
                                test_parallel_count=test_parallel_count, \
                                checkpoint_handler=checkpoint_handler, \
                                cp_calling_func_name=cp_func_name, \
                                cp_calling_done_task_id=(cp_task_id - 1), \
                                cp_calling_tool=criterion.get_str())

                # @Checkpoint: checkpoint
                checkpoint_handler.do_checkpoint(func_name=cp_func_name, \
                                taskid=cp_task_id, tool=criterion.get_str())

        # @Checkpoint: Finished (for time)
        checkpoint_handler.set_finished(None)
    #~ def runtests_criteria_coverage()

    @staticmethod
    def _extract_sub_dicts(dict_obj, in_key_list_list):
        ERROR_HANDLER.assert_true(len(in_key_list_list) > 0, \
                                                    "empty key list", __file__)
        out_dict_list = []
        for key_list in in_key_list_list:
            out_dict_list.append({k: dict_obj[k] for k in key_list})
        return out_dict_list
    #~ def _extract_sub_dicts()

    @staticmethod
    def _get_criteria_groups(criterion2executable_path, \
                                                 criterion2environment_vars):
        """ separate to common executions 
         (those with same exe_path_map and env_vars)
        :return: list of groups, each group info in the tuple
        """
        ERROR_HANDLER.assert_true(set(criterion2executable_path) == \
                        set(criterion2environment_vars), "Missmatch", __file__)
        criterialist = list(criterion2executable_path.keys())
        groups = []
        visited = set()
        for c_pos, criterion in enumerate(criterialist):
            if criterion not in visited:
                # add its group
                groups.append(([criterion], \
                                    criterion2executable_path[criterion],
                                    criterion2environment_vars[criterion]))
                visited.add(criterion)
                # add anything else from same group
                for e_pos in range(c_pos+1, len(criterialist)):
                    if criterion2executable_path[criterialist[e_pos]] == \
                                            groups[-1][1] and \
                            criterion2environment_vars[criterialist[e_pos]] ==\
                                                        groups[-1][2]:
                       groups[-1][0].append(criterialist[e_pos]) 
                       visited.add(criterialist[e_pos])
        return groups
    #~ def _get_criteria_groups()

    def instrument_code (self, enabled_criteria, exe_path_map=None, \
                                code_builds_factory_override=None, \
                                parallel_count=1):
        '''
            (TODO: support parallelism: per test outdata)
        '''

        logging.debug("# Instrumenting code with {} ...".format(\
                                        self.config.get_tool_config_alias()))

        # FIXME: Support parallelism, then remove the code
        # bellow:
        ERROR_HANDLER.assert_true(parallel_count <= 1, \
                "FIXME: Must first implement support for parallel mutatio")
        #~ FXIMEnd

        # @Checkpoint: create a checkpoint handler (for time)
        checkpoint_handler = CheckPointHandler(self.get_checkpointer())
        if not checkpoint_handler.is_finished():

            ERROR_HANDLER.assert_true(len(enabled_criteria) > 0, \
                                        "no criterion is enabled", __file__)

            if code_builds_factory_override is None:
                code_builds_factory_override = self.code_builds_factory
            
            if exe_path_map is None:
                exe_path_map = code_builds_factory_override.repository_manager\
                                                            .get_exe_path_map()

            self._do_instrument_code ( \
                            exe_path_map=exe_path_map, \
                            code_builds_factory=code_builds_factory_override, \
                            enabled_criteria=enabled_criteria, \
                            parallel_count=parallel_count)

            # @Checkpoint: Finished (for time)
            checkpoint_handler.set_finished(None)
    #~ def instrument_code()

    def _get_latest_top_output_dir(self):
        return os.path.dirname(os.path.dirname(self.criteria_working_dir))
    #~ def _get_latest_top_output_dir()

    def _extract_metaoutlog_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        """ Override this in a tool that suports it
            By default it is not supported and thus fails
        """
        ERROR_HANDLER.assert_true("meta outlog unsupported by this tool")
        return {}
    #~ def _extract_metaoutlog_data_of_a_test()
        
    def tool_installed (self):
        """ Check that a tool with given conf is installed
        """
        return self.installed(custom_binary_dir=self.custom_binary_dir)
    #~ def tool_installed ()

    def _dir_chmod777(self, dirpath):
        try:
            for root_, dirs_, files_ in os.walk(dirpath):
                for sub_d in dirs_:
                    os.chmod(os.path.join(root_, sub_d), 0o777)
                for f_ in files_:
                    os.chmod(os.path.join(root_, f_), 0o777)
        except PermissionError:
            ret,_,_ = DriversUtils.execute_and_get_retcode_out_err('sudo', \
                                        ['chmod 777 -R {}'.format(dirpath)])
            ERROR_HANDLER.assert_true(ret == 0, \
                        "'sudo chmod 777 -R "+dirpath+"' failed (returned "+\
                                                        str(ret)+")", __file__)
    #~ def _dir_chmod777()

    #######################################################################
    ##################### Methods to implement ############################
    #######################################################################

    #@abc.abstractclassmethod
    @classmethod
    @abc.abstractclassmethod
    def installed(cls, custom_binary_dir=None):
        """ Check that the tool is installed
            :return: bool reprenting whether the tool is installed or not 
                    (executable accessible on the path)
                    - True: the tool is installed and works
                    - False: the tool is not installed or do not work
        """
        print ("!!! Must be implemented in child class !!!")
    #~ def installed()

    @classmethod
    @abc.abstractclassmethod
    def _get_meta_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in same file
            :return: list of citeria
        """
        #return [
        #        TestCriteria.STATEMENT_COVERAGE,
        #        TestCriteria.BRANCH_COVERAGE,
        #        TestCriteria.FUNCTION_COVERAGE,
        #       ]
        print ("!!! Must be implemented in child class !!!")
    #~ def _get_meta_instrumentation_criteria()

    @classmethod
    @abc.abstractclassmethod
    def _get_separated_instrumentation_criteria(cls):
        """ Criteria where all elements are instrumented in different files
            :return: list of citeria
        """
        #return [TestCriteria.STRONG_MUTATION]
        print ("!!! Must be implemented in child class !!!")
    #~ def _get_separated_instrumentation_criteria()

    @abc.abstractmethod
    def get_instrumented_executable_paths_map(self, enabled_criteria):
        print ("!!! Must be implemented in child class !!!")
    #~ def get_instrumented_executable_paths_map()

    @abc.abstractmethod
    def get_criterion_info_object(self, criterion):
        print ("!!! Must be implemented in child class !!!")
    #~ def get_criterion_info_object(self, criterion)

    @abc.abstractmethod
    def _get_criterion_element_executable_path(self, criterion, element_id):
        print ("!!! Must be implemented in child class !!!")
    #~ def _get_criterion_element_executable_path

    @abc.abstractmethod
    def _get_criterion_element_environment_vars(self, criterion, element_id):
        '''
            return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        print ("!!! Must be implemented in child class !!!")
    #~ def _get_criterion_element_environment_vars()

    @abc.abstractmethod
    def _get_criteria_environment_vars(self, result_dir_tmp, enabled_criteria):
        '''
        return: python dictionary with environment variable as key
                     and their values as value (all strings)
        '''
        print ("!!! Must be implemented in child class !!!")
    #~ def _get_criteria_environment_vars()

    @abc.abstractmethod
    def _collect_temporary_coverage_data(self, criteria_name_list, \
                                            test_execution_verdict, \
                                            used_environment_vars, \
                                            result_dir_tmp, \
                                            testcase):
        '''
        '''
        print ("!!! Must be implemented in child class !!!")
    #~ def _collect_temporary_coverage_data()

    @abc.abstractmethod
    def _extract_coverage_data_of_a_test(self, enabled_criteria, \
                                    test_execution_verdict, result_dir_tmp):
        '''
            return: the dict of statements with covering count
        '''
        print ("!!! Must be implemented in child class !!!")
    #~ def _extract_coverage_data_of_a_test()

    @abc.abstractmethod
    def _do_instrument_code (self, exe_path_map, code_builds_factory, \
                                        enabled_criteria, parallel_count=1):
        print ("!!! Must be implemented in child class !!!")
    #~ def _do_instrument_code()
#~ class BaseCriteriaTool



import os
import sys
from muteria.common.mix import GlobalConstants
from muteria.drivers import DriversUtils

def python_unittest_runner(test_name, repo_root_dir, exe_path_map, env_vars, \
                                            timeout, collected_output=None):
    # TODO: use exe_path_map 

    def parse_test(s):
        return s.split('...')[0].replace(':','/').replace(' ','')

    cwd = os.getcwd()
    os.chdir(repo_root_dir)

    try:
        args_list = ['-m', 'unittest', test_name, '-v']

        if collected_output is None:
            retcode, stdout, _ = DriversUtils.execute_and_get_retcode_out_err(\
                                prog=sys.executable, args_list=args_list, \
                                        timeout=timeout, merge_err_to_out=True)
            stdout = stdout.splitlines()
        else:
            # collected_output is a list ([retcode, out_err_log])
            # TODO: use wrapper? or parse out for particular codes... 
            assert False, "TO BE Implemented"
    except:
        # ERROR
        os.chdir(cwd)
        return GlobalConstants.TEST_EXECUTION_ERROR
    
    # Parse the result
    subtests_verdicts = {}
    hasfail = False
    hasfail |= (retcode != 0)
    for s in stdout:
        if s.endswith('... FAIL'):
            hasfail = True
            subtests_verdicts[parse_test(s)] = True
        elif s.endswith('... ok'):
            subtests_verdicts[parse_test(s)] = False
    #print(subtests_verdicts)
    os.chdir(cwd)
    return GlobalConstants.FAIL_TEST_VERDICT if hasfail else \
                                            GlobalConstants.PASS_TEST_VERDICT
#~ def python_unittest_runner()


from muteria.drivers.testgeneration.tools_by_languages.c.shadow_se.shadow_se \
                                                import TestcasesToolShadowSE 
StaticTestcaseTool = None
DynamicTestcaseTool = TestcasesToolShadowSE


""" Defaults parameters that are common to all languages
"""
from __future__ import print_function

from muteria.configmanager.configurations import SessionMode
from muteria.configmanager.configurations import TestcaseToolsConfig
from muteria.configmanager.configurations import CriteriaToolsConfig
from muteria.configmanager.configurations import ToolUserCustom

from muteria.drivers.criteria import TestCriteria
from muteria.drivers.criteria import CriteriaToolType

from muteria.drivers.testgeneration import TestToolType


#######################################################
#######             Execution Parameters         ######
#######################################################
# Criteria to Include in the analysis
# The corresponding configs of codecoverage (`CodecoverageToolsConfig`)
# and mutation (`MutationToolsConfig`) are considered if and only if
# the corresponding include is enabled

# Decide whether to start over, deleting previous, for execution mode
EXECUTION_CLEANSTART=False

# Value of type SessionMode (Mandatory)
RUN_MODE = None

# List of enabled criteria
# (must have tool specified is specifically enabled)
# None mean all criteria with tool specified
ENABLED_CRITERIA = []

# Enable keeping output summary for program passfail execution
GET_PASSFAIL_OUTPUT_SUMMARY = True

# keepoutput summary for the following criteria (may run slower)
CRITERIA_WITH_OUTPUT_SUMMARY = []

# Decides whether to hash the output log
HASH_OUTLOG = True

# PARALELISM
SINGLE_REPO_PARALLELISM = 1 # Max number of parallel exec in a repo dir

# MICRO CONTROLS
EXECUTE_ONLY_CURENT_CHECKPOINT_META_TASK = False # for Debugging
RESTART_CURRENT_EXECUTING_META_TASKS = False
# Specify a Step to go back to
RE_EXECUTE_FROM_CHECKPOINT_META_TASKS = [] # Make interaction easy

# Output dir pathname (Mandatory)
OUTPUT_ROOT_DIR = None

# Enable logging debug data
LOG_DEBUG = False

#######################################################
#######             Reporting Parameters         ######
#######################################################
GENERATE_HTML_REPORT = True

OUTPUT_CRITERIA_SCORES = True
CRITERIA_SCORE_BY_TOOL = True
OUTPUT_CRITERIA_SUBSUMPTION_SCORE = True

OUTPUT_CRITERIA_COVERED_ELEMENTS = False
OUTPUT_CRITERIA_UNCOVERED_ELEMENTS = True
DETAILED_ELEMENTS_OUTPUT = False

OUTPUT_CRITERIA_SUBSUMING_ELEM_NUM = True

# When iteratively try to cover some element, show
OUTPUT_STATS_HISTORY = True

#######################################################
#######        Project Config Parameters         ######
#######################################################
# Project programming language (Mandatory)
PROGRAMMING_LANGUAGE = None

# Repository dir pathname (Mandatory)
REPOSITORY_ROOT_DIR = None

# string representing the relative path to the executable
# (or entry point file) in the repository
REPO_EXECUTABLE_RELATIVE_PATHS = None

# optional map between each source and the corresponding intermediate
# file map (such as object or assembly file for c of .class file for java)
TARGET_SOURCE_INTERMEDIATE_CODE_MAP = None

# Specify the optional general scope of the evaluation,
# Specific scope can be specified per tool
TARGET_CLASSES_NAMES = None
# None value mean all functions
TARGET_METHODS_BY_TARGET_CLASSES = None
# each source file element is the relative path from repos rootdir.
# None value means all source files
##TARGET_SOURCE_FILES_LIST = None

# each test file element is the relative path from repos rootdir.
# None value means all dev tests
DEVELOPER_TESTS_LIST = None

# Function that takes 3 arguments:
#   <test_name: str>
#   <repos directory rootdir: str>
#   <Executable relative path map: dict>
#   <env_vars: map>
#   <timeout: int>
# and run with the executable as in repository
# The function return:
#   0 on passing test
#   1 on failing test
#   -1 on error
CUSTOM_DEV_TEST_RUNNER_FUNCTION = None

CUSTOM_DEV_TEST_PROGRAM_WRAPPER_CLASS = None

# Take the outlog as a string and return the cleaned version
CUSTOM_TEST_EXECUTION_OUTPUT_CLEANER_FUNCTION = None

# Optional. When not None, the CUSTOM_DEV_TEST_RUNNER is the name of
# the function in this file to use
CUSTOM_DEV_TEST_RUNNER_MODULE = None

# Function that build the code to execute
# (for compiled languages such as C)
# The function has 5 parameters:
#   <repos directory rootdir: str>
#   <Executable relative path: str>
#   <Optional compiler to use(compiler name): str>
#   <optional flags to pass to compiler: list of flags>
#   <clean temporary before build: bool>
#   <reconfigure before build: bool>
# And returns:
#   True on success
#   False on failure
# A code builder class is defined and make use of this function
# It ensure on call to this functin at the time and no call while any
# processing in the repodir.
CODE_BUILDER_FUNCTION = None

# Optional. When not None, the CODE_BUILDER_FUNC is the name of
# the function in this file to use
CODE_BUILDER_MODULE = None

#######################################################
#######             Tools parameters             ######
#######################################################
# ===================== TESTCASES =====================#
# Tests already existing before execution starts
DEVELOPER_TESTS_ENABLED = True
# Test created during execution
GENERATED_TESTS_ENABLED = True

STOP_TESTS_EXECUTION_ON_FAILURE = False # Will not get full matrix

DISCARD_FLAKY_TESTS = True

# Test tool types in order of execution (Test generation).
# elements in same tuple have parallel execution. elements at tuple i
# generate tests after element at i-1 and may use results of test from i-1.
TEST_TOOL_TYPES_SCHEDULING = [
(TestToolType.USE_ONLY_CODE,),
(TestToolType.USE_CODE_AND_TESTS,),
]

# Test generation tool. Example:
# >>> TestcaseToolsConfig(tooltype=TestToolType.USE_ONLY_CODE,
#                        toolname='klee', config_id=0)
TESTCASE_TOOLS_CONFIGS = [

]

TESTCASES_SELECTION = None

# Reporting
REPORT_NUMBER_OF_TESTS_GENERATED = True
REPORT_NUMBER_OF_DUPLICATED_TESTS = True

## --- Modifiable (Testcase) ---##
# use test case oracle as oracle
#TESTS_ORACLE_TESTS = True
# Use output of the specified version as oracle,
# Pass filepath to repo patch
#TESTS_ORACLE_OTHER_VERSION = None
# file path to an executable to use as oracle
#TESTS_ORACLE_OTHER_EXECUTABLE = None

#TEST_GENERATION_TIMEOUT = 7200.0 # in seconds
#ONE_TEST_EXECUTION_TIMEOUT = 900.0 # in seconds (Handle inifnite loops)
# ========================================================#

# ===================== CRITERIA COVERAGE =====================#

# Map with key criteria and values the list of tools
# criteria tool. Example:
# >>> CriteriaToolConfig(tooltype=CriteriaToolType.USE_ONLY_CODE,
#                        toolname='gcov', config_id=0)
CRITERIA_TOOLS_CONFIGS_BY_CRITERIA = {

}

# List of sets of test criteria stating the order in which criteria
# should be executed. Example strom mutation after weak mutation.
# When None, Default to the order in:
# >>> muteria.drivers.criteria.CRITERIA_SEQUENCE
CRITERIA_SEQUENCE = None

# List of criteria that have test objectives covered when test execution
# differs with original
# When None, Default to the order in:
# >>> muteria.drivers.criteria.CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM
CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM = None

# List of criteria to run with failing tests
RUN_FAILING_TESTS_WITH_CRITERIA = [

]

# List of criteria to run with passing tests
RUN_PASSING_TESTS_WITH_CRITERIA = [

]

CRITERIA_RESTRICTION_ENABLED = True  # Enable restricting mutation(scope)

# criterion: selection tools. Example: SM and TCE or E-SELECTIVE
CRITERIA_ELEM_SELECTIONS = {

}

ONLY_EXECUTE_SELECTED_CRITERIA_ELEM = True

MAX_CRITERIA_ELEM_SELECTION_NUM_PERCENT = '100%'

# Criterion: guider dict. ex: {STRONG_MUTATION: Surviving}
CRITERIA_TESTGEN_GUIDANCE = {

}

# Criterion: optimizers dict. ex: {STRONG_MUTATION: SM_OPTIMIZED_BY_WM}
CRITERIA_EXECUTION_OPTIMIZERS = {

}

# Will not get full matrix. the non executed will be uncertain
COVER_CRITERIA_ELEMENTS_ONCE = False

## --- Modifiable (Code) ---##
#SEPARATED_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 60.0 # in seconds
#META_TEST_EXECUTION_EXTRA_TIMEOUT_TIMES = 600.0 # in seconds
# ========================================================#

#######################################################
#######             Extra parameters             ######
#######################################################
#LLVM_TO_NATIVE_LINKING_FLAGS = None



#
# [LICENCE]
#
""" From Json file criteria test execution optimizer class
"""

from __future__ import print_function
import os
import sys
import copy
import logging

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.drivers.optimizers.criteriatestexecution.tools.fromdict \
                            import CriteriaTestExecutionOptimizer as DictOpt

from muteria.drivers.optimizers.testexecution.tools.default \
                                                import TestExecutionOptimizer

ERROR_HANDLER = common_mix.ErrorHandler

class CriteriaTestExecutionOptimizer(DictOpt):
    def __init__(self, config, explorer, criterion, jsonfile=None):
        ERROR_HANDLER.assert_true(jsonfile is not None, \
                                        "jsonfile cannot be None", __file__)
        jsonobj = common_fs.loadJSON(jsonfile)
        DictOpt.__init__(self, config, explorer, criterion, dictobj=jsonobj)
    #~ def __init__()
#~ class CriteriaTestExecutionOptimizer




"""
TODO: Add support for multi versions of results 
"""

from __future__ import print_function

import os 
import logging
import shutil
import glob
import math
import copy
import random
import inspect

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

from muteria.repositoryandcode.repository_manager import RepositoryManager
from muteria.repositoryandcode.code_builds_factory import CodeBuildsFactory

from muteria.drivers import DriversUtils
from muteria.drivers.testgeneration import TestToolType
from muteria.drivers.testgeneration.meta_testcasetool import MetaTestcaseTool
import muteria.drivers.criteria as criteria_pkg
from muteria.drivers.criteria.meta_testcriteriatool import MetaCriteriaTool

import muteria.drivers.optimizers.criteriatestexecution.optimizerdefs as \
                                                                crit_opt_module

from muteria.statistics.main import StatsComputer

import muteria.controller.explorer as outdir_struct
import muteria.controller.logging_setup as logging_setup
import muteria.controller.checkpoint_tasks as checkpoint_tasks

ERROR_HANDLER = common_mix.ErrorHandler

class CheckpointData(dict):
    """
    """
    def __init__(self):
        pass
    #~ def __init__()

    def update_all(self, tasks_obj, test_types, test_types_pos, \
                                criteria_set, criteria_set_pos):
        self.update_tasks_obj(tasks_obj)
        self.update_test_types(test_types_pos, test_types)
        self.update_criteria_set(criteria_set_pos, criteria_set)
    #~ def update_all()

    def get_json_obj(self):
        retval = dict(self.__dict__)
        retval['tasks_obj'] = retval['tasks_obj'].get_as_json_object()
        retval['test_types'] = [tt.get_str() for tt in retval['test_types']]
        if self.criteria_set is not None:
            retval['criteria_set'] = \
                                [c.get_str() for c in retval['criteria_set']]
        return retval
    #~ def get_json_obj()

    def update_from_json_obj(self, json_obj):
        self.update_all(**json_obj)
        self.tasks_obj = checkpoint_tasks.TaskOrderingDependency(\
                                                    json_obj=self.tasks_obj)
        self.test_types = tuple([TestToolType[tt] for tt in self.test_types])
        if self.criteria_set is not None:
            self.criteria_set = set(\
                    [criteria_pkg.TestCriteria[c] for c in self.criteria_set])
    #~ def update_from_json_obj()

    def test_tool_types_is_executed(self, seq_id, test_tool_types):
        if self.test_types_pos > seq_id:
            return True
        if self.test_types_pos == seq_id and \
                                set(self.test_types) != set(test_tool_types):
            ERROR_HANDLER.error_exit("cp test_tool_type changed", __file__)
        return False
    #~ def test_tool_types_is_executed()

    def switchto_new_test_tool_types(self, seq_id, test_tool_types):
        if seq_id > self.test_types_pos:
            self.tasks_obj.set_task_back_as_todo_executing(\
                            checkpoint_tasks.Tasks.TESTS_GENERATION_GUIDANCE)
            self.update_test_types(seq_id, test_tool_types)
            self.update_criteria_set(None, None)
    #~ def switchto_new_test_tool_types()

    def criteria_set_is_executed(self, seq_id, criteria_set):
        if self.criteria_set_pos is not None:
            if self.criteria_set_pos > seq_id:
                return True
            if self.criteria_set_pos == seq_id:
                ERROR_HANDLER.assert_true(set(self.criteria_set) \
                                                        == set(criteria_set), \
                                        "cp criteria_set changed", __file__)
                return True
        return False
    #~ def criteria_set_is_executed()

    def switchto_new_criteria_set(self, seq_id, criteria_set):
        if self.criteria_set_pos is None or seq_id > self.criteria_set_pos:
            self.update_criteria_set(seq_id, criteria_set)
    #~ def switchto_new_criteria_set()

    def update_tasks_obj(self, tasks_obj):
        self.tasks_obj = tasks_obj
    #~ def update_tasks_obj()

    def update_test_types(self, test_types_pos, test_types):
        self.test_types = test_types
        self.test_types_pos = test_types_pos
    #~ def update_test_types()
        
    def update_criteria_set(self, criteria_set_pos, criteria_set):
        self.criteria_set = criteria_set
        self.criteria_set_pos = criteria_set_pos
    #~ def update_criteria_set()
#~ class CheckpointData

class Executor(object):
    """ Execution Orchestration class
        The execution entry point is the method 'main' 
    """

    def __init__(self, config, top_timeline_explorer):
        """ The various configurations for the execution are passed here, as
            well as the corresponding directory structure
        """
        self.config = config
        self.top_timeline_explorer = top_timeline_explorer

        self.head_explorer = self.top_timeline_explorer.get_latest_explorer()
        # Initialize output structure
        self._initialize_output_structure(cleanstart=\
                                self.config.EXECUTION_CLEANSTART.get_val())
        if not logging_setup.is_setup():
            if self.config.LOG_DEBUG.get_val():
                logging_setup.setup(file_level=logging.DEBUG, \
                            console_level=logging.DEBUG, \
                            logfile=self.head_explorer.get_file_pathname(\
                                                outdir_struct.MAIN_LOG_FILE))
                logging.debug("Logging Debug Level")
            else:
                logging_setup.setup(\
                            logfile=self.head_explorer.get_file_pathname(\
                                                outdir_struct.MAIN_LOG_FILE))
        # Create repo manager
        # XXX The repo manager automatically revert any previous problem
        self.repo_mgr = Executor.create_repo_manager(config)

        # set error handlers for revert repo
        common_mix.ErrorHandler.set_corresponding_repos_manager(self.repo_mgr)

        # create codes builders
        self.cb_factory = CodeBuildsFactory(self.repo_mgr, \
                                workdir=self.head_explorer.get_file_pathname(\
                                            outdir_struct.CB_FACTORY_WORKDIR))
    #~ def __init__()

    def main(self):
        """ Executor entry point
        """
        # Initialize output structure (done in __init__)
        #self._initialize_output_structure(\
        #                cleanstart=self.config.EXECUTION_CLEANSTART.get_val())

        # Make checkpointer
        self.checkpointer = \
                    common_fs.CheckpointState(*self._get_checkpoint_files())
        was_finished = False
        if self.checkpointer.is_finished():
            if len(self.config.RE_EXECUTE_FROM_CHECKPOINT_META_TASKS.\
                                                                get_val()) > 0:
                self.checkpointer.restart_task()
                was_finished = True
            else:
                return

        # Meta testcases tool
        self.meta_testcase_tool = self._create_meta_test_tool(self.config, \
                                                            self.head_explorer)
        self.meta_testcase_tool.check_tools_installed()

        # Meta criteria
        self.meta_criteria_tool = self._create_meta_criteria_tool(self.config,\
                                                    self.meta_testcase_tool)
        self.meta_criteria_tool.check_tools_installed()

        # Test generation guidance
        self.meta_testgen_guidance_tool = self._create_meta_testgen_guidance(\
                                                                self.config)

        # Test execution optimization
        self.meta_testexec_optimization_tool = \
                        self._create_meta_testexec_optimization(self.config)

        # Criteria generation guidance
        self.meta_criteriagen_guidance_tool = \
                            self._create_meta_criteriagen_guidance(self.config)

        # Criteria optimization
        self.meta_criteriaexec_optimization_tools = \
                    self._create_meta_criteriaexec_optimization(self.config)

        # See whether starting or continuing
        check_pt_obj = self.checkpointer.load_checkpoint_or_start(\
                                            ret_detailed_exectime_obj=False)
        
        test_tool_type_sequence = \
                            self.config.TEST_TOOL_TYPES_SCHEDULING.get_val()

        # Do nothing if no tool types specified
        if len(test_tool_type_sequence) == 0:
            return
            
        self.cp_data = CheckpointData()
        if check_pt_obj is not None:
            # Continuing
            self.cp_data.update_from_json_obj(json_obj=check_pt_obj)
        else:
            # Staring anew
            self.cp_data = CheckpointData()
            self.cp_data.update_all(\
                        tasks_obj=checkpoint_tasks.TaskOrderingDependency(),\
                        test_types=test_tool_type_sequence[0],\
                        test_types_pos=0,\
                        criteria_set=None,\
                        criteria_set_pos=None)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

        # Ensure that the repository exe and obj are in default state
        self.cb_factory.set_repo_to_build_default()

        # Execute the tasks
        task_set = None
        # by testtool type sequence loop
        for seq_id, test_tool_types in enumerate(test_tool_type_sequence):
            
            logging.info("")
            logging.info("executor: Running for Test tool types = "+\
                                str([t.get_str() for t in test_tool_types]))
            logging.info("")

            # Was it already checkpointed w.r.t test type seq
            if self.cp_data.test_tool_types_is_executed(\
                                                    seq_id, test_tool_types):
                continue

            #if len(self.meta_testcase_tool.get_candidate_tools_aliases(\
            #                        test_tool_type_list=test_tool_types)) == 0:
            #   continue
            
            # If we have a new seq_id, it is another test type loop
            # NOTE: This switch is written at following steps (tasks...)
            self.cp_data.switchto_new_test_tool_types(seq_id, test_tool_types)

            # Actual Execution in the temporary dir (until before finish)
            while True:
                # (Break flow)
                if len(self.config.RE_EXECUTE_FROM_CHECKPOINT_META_TASKS.\
                                                                get_val()) > 0:
                    if was_finished:
                        # XXX: go to the state right before finished is set
                        self.cp_data.tasks_obj.set_all_tasks_to_completed()
                    for task in self.config\
                                        .RE_EXECUTE_FROM_CHECKPOINT_META_TASKS\
                                        .get_val():
                        self.cp_data.tasks_obj\
                                        .set_task_back_as_todo_untouched(task)
                    self.config.RE_EXECUTE_FROM_CHECKPOINT_META_TASKS\
                                                                .set_val([])

                # 1. get executing task
                task_set = self.cp_data.tasks_obj.get_next_todo_tasks()

                # 2. stop if the next task is FINISHED
                if len(task_set) == 0 or (len(task_set) == 1 and \
                        list(task_set)[0] == checkpoint_tasks.Tasks.FINISHED):
                    break

                # 3. stop if the next task is AGGREGATED_STATS and there 
                # is still iteration from test_tool_type_sequence
                if (len(task_set) == 1 \
                                and list(task_set)[0] == \
                                    checkpoint_tasks.Tasks.AGGREGATED_STATS \
                                and seq_id < len(test_tool_type_sequence) - 1):
                    break

                # (Break flow)
                # use the rerun from here, only here, stop after here
                if self.config.RESTART_CURRENT_EXECUTING_META_TASKS.get_val():
                    # restarts the tasks
                    for task in task_set:
                        self.cp_data.tasks_obj.\
                                        set_task_back_as_todo_untouched(task)
                    # disable it
                    self.config.RESTART_CURRENT_EXECUTING_META_TASKS.set_val(\
                                                                        False)
                    
                
                # 3. execute the tasks  (TODO: parallelism)
                for task in task_set:
                    self._execute_task(task)

                # (Break flow)
                if self.config.EXECUTE_ONLY_CURENT_CHECKPOINT_META_TASK.\
                                                                    get_val():
                    ERROR_HANDLER.error_exit("(DBG) Stopped after executing"
                                " only the current checkpoint meta tasks."
                                " because of corresponding variable was set", \
                                                                    __file__)

        ERROR_HANDLER.assert_true(task_set is not None, \
                "There must be task set (empty list or FINISHED)", __file__)

        # Set checkpoint to finished
        if len(task_set) == 1:
            ERROR_HANDLER.assert_true(list(task_set)[0] == \
                                            checkpoint_tasks.Tasks.FINISHED,\
                                        "task must be Finished Here", __file__)
            self.cp_data.tasks_obj.set_task_completed(list(task_set)[0])
        else:
            ERROR_HANDLER.assert_true(len(task_set) == 0, \
                                    "task set must be empty here", __file__)

        self.checkpointer.set_finished()
    #~ def main()

    def get_repo_manager(self):
        return self.repo_mgr
    #~ def get_repo_manager()

    def custom_execution(self):
        """ Interact with the user and ask what to do.
            For execution of tests with custom executable for example
        """
        # XXX: check that the normal execution was finished (from checkpointer)
        self.checkpointer = \
                    common_fs.CheckpointState(*self._get_checkpoint_files())
        ERROR_HANDLER.assert_true(self.checkpointer.is_finished(), \
                        "custom execution must be ran when run is finished", \
                                                                    __file__)

        mode = None
        while mode not in ['tests', 'crtiteria_tests']:
            mode = input (
                    "> Chose what to execute ('tests' or 'criteria_tests'): ")
        custom_out = input("> Input the non existing custom output dir: ")
        ERROR_HANDLER.assert_true(not os.path.isdir(custom_out), \
                                            "custom_out is existing", __file__)
        ERROR_HANDLER.assert_true(os.path.isdir(os.path.dirname(custom_out)), \
                            "parent of custom_out is not existing", __file__)
        if mode == 'tests':
            ## read exe
            exe_map_str = input("> Input existing executable file map to use"
                        " ({<path in repo>: <used exe path>}): ")
            exe_map = common_fs.json.loads(exe_map_str)
            ERROR_HANDLER.assert_true(type(exe_map) == dict, \
                                                "invalid path map", __file__)
            for kp, rp in exe_map.items():
                ERROR_HANDLER.assert_true(os.path.isfile(rp), \
                            "executable is missing ({}: {})".format(kp, rp), \
                                                                    __file__)
            ## read test list
            testlist_file = input("> Input existing test list file to use: ")
            ERROR_HANDLER.assert_true(os.path.isfile(testlist_file), \
                    "test file list not existing ({})".format(testlist_file),\
                                                                    __file__)
            with open(testlist_file) as f:
                testlist = [t.strip() for t in f]

            # XXX execution
            ## Create meta test tool
            try:
                meta_testcase_tool = self.meta_testcase_tool
            except AttributeError:
                meta_testcase_tool = self._create_meta_test_tool(self.config, \
                                                            self.head_explorer)
                meta_testcase_tool.check_tools_installed()
            ## create test prioritization tool
            try:
                meta_testexec_optimization_tool = \
                                        self.meta_testexec_optimization_tool
            except AttributeError:
                meta_testexec_optimization_tool = \
                        self._create_meta_testexec_optimization(self.config)

            ## Create outdir
            os.mkdir(custom_out)

            ## run the tests

            matrix_file_key = outdir_struct.TEST_PASS_FAIL_MATRIX
            matrix_file = os.path.join(custom_out, matrix_file_key)
            execoutput_file_key = \
                                outdir_struct.PROGRAM_TESTEXECUTION_OUTPUT
            execoutput_file = None
            if self.config.GET_PASSFAIL_OUTPUT_SUMMARY.get_val():
                execoutput_file = os.path.join(custom_out, execoutput_file_key)

            self.meta_testcase_tool.runtests(meta_testcases=testlist, \
                        exe_path_map=exe_map,
                        stop_on_failure=\
                                self.config.STOP_TESTS_EXECUTION_ON_FAILURE\
                                                                .get_val(), \
                        recalculate_execution_times=False, \
                        fault_test_execution_matrix_file=matrix_file, \
                        fault_test_execution_execoutput_file=execoutput_file, \
                        test_prioritization_module=\
                                        meta_testexec_optimization_tool, \
                        finish_destroy_checkpointer=True)
        elif mode == 'criteria_tests':
            ERROR_HANDLER.error_exit("To be implemented")
        else:
            ERROR_HANDLER.error_exit("invalid mode and bug", __file__)
    #~ def custom_execution()

    def _execute_task(self, task):
        """ TODO: 
                1. implement the todos here
        """
        
        # @Checkpointing: see whether the task is untouch to clear tmp
        task_untouched = self.cp_data.tasks_obj.task_is_untouched(task)

        # Execute base on the task: ---------------------------------------
        if task == checkpoint_tasks.Tasks.STARTING:
            pass
        elif task == checkpoint_tasks.Tasks.FINISHED:
            pass

        elif task == checkpoint_tasks.Tasks.TESTS_GENERATION_GUIDANCE:
            pass #TODO: Generate focus criteria and tests and store in json file
        elif task == checkpoint_tasks.Tasks.TESTS_GENERATION:
            # @Checkpointing
            if task_untouched:
                #if self.cp_data.test_types_pos == 0:
                #    self.meta_testcase_tool.clear_working_dir() 
                if self.meta_testcase_tool.has_checkpointer():
                    self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # Generate the tests without criteria instrumented
            self.meta_testcase_tool.generate_tests(\
                            meta_criteria_tool_obj=None, \
                            test_tool_type_list=self.cp_data.test_types)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())
            # Destroy meta test checkpointer
            self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

        elif task == checkpoint_tasks.Tasks.TESTS_GENERATION_USING_CRITERIA:
            # @Checkpointing
            if task_untouched:
                #if self.cp_data.test_types_pos == 0:
                #    self.meta_testcase_tool.clear_working_dir() 
                if self.meta_testcase_tool.has_checkpointer():
                    self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # Generate the tests using criteria
            self.meta_testcase_tool.generate_tests(\
                            meta_criteria_tool_obj=self.meta_criteria_tool, \
                            test_tool_type_list=self.cp_data.test_types)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())
            # Destroy meta test checkpointer
            self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

        elif task == checkpoint_tasks.Tasks.\
                                    TESTS_EXECUTION_SELECTION_PRIORITIZATION:
            out_file_key = outdir_struct.TMP_SELECTED_TESTS_LIST
            out_file = self.head_explorer.get_file_pathname(out_file_key)
                                        
            # @Checkpointing
            if task_untouched:
                if self.meta_testcase_tool.has_checkpointer():
                    self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()
                self.head_explorer.remove_file_and_get(out_file_key)
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            candidate_aliases = \
                        self.meta_testcase_tool.get_candidate_tools_aliases(\
                                test_tool_type_list=self.cp_data.test_types)

            all_tests = self.meta_testcase_tool.get_testcase_info_object(\
                                   candidate_tool_aliases=candidate_aliases)\
                                                            .get_tests_list()

            selected_tests = []
            for meta_test in all_tests:
                toolalias, _ = DriversUtils.reverse_meta_element(meta_test)
                if toolalias in candidate_aliases:
                    selected_tests.append(meta_test)

            # XXX use the actual selector. For now just use all tests
            sel_tech = self.config.TESTCASES_SELECTION.get_val()
            # selection method
            if inspect.isfunction(sel_tech):
                ERROR_HANDLER.assert_true(\
                                    len(inspect.getargspec().args) == 2, \
                                    "Test selection function must take 2 args "
                                    "(testlist, maxselcount)", __file__)
                
            else:
                if sel_tech is not None and sel_tech != 'DummyRandom':
                    ERROR_HANDLER.error_exit(\
                        'only random tests selection supported now!', \
                                                                __file__)
                # make DummyRandom selection
                sel_tech = random.sample

            # make selection
            selected_tests = sel_tech(selected_tests, len(selected_tests))

            # Check for flakiness
            logging.debug("# Checking for tests flakiness ...")
            flaky_tests = self.meta_testcase_tool.check_get_flakiness(\
                                                                selected_tests)
            if len(flaky_tests) > 0:
                if self.config.DISCARD_FLAKY_TESTS.get_val():
                    selected_tests = \
                                list(set(selected_tests) - set(flaky_tests))
                else:
                    ERROR_HANDLER.error_exit("There are Flaky tests!!", \
                                                                    __file__)
            else:
                logging.debug("# No Flaky Test :)")

            # write the tests
            common_fs.dumpJSON(list(selected_tests), out_file)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())
            # Destroy meta test checkpointer
            #self.meta_testexec_optimization_tool.get_checkpoint_state_object()\
            #                                            .destroy_checkpoint()
        elif task == checkpoint_tasks.Tasks.PASS_FAIL_TESTS_EXECUTION:
            # Make sure that the Matrices dir exists
            self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.RESULTS_MATRICES_DIR)

            matrix_file_key = outdir_struct.TMP_TEST_PASS_FAIL_MATRIX
            matrix_file = self.head_explorer.get_file_pathname(matrix_file_key)
            execoutput_file_key = \
                                outdir_struct.TMP_PROGRAM_TESTEXECUTION_OUTPUT
            if self.config.GET_PASSFAIL_OUTPUT_SUMMARY.get_val():
                # Make sure that the exec Output dir exists
                self.head_explorer.get_or_create_and_get_dir(\
                            outdir_struct.RESULTS_TESTEXECUTION_OUTPUTS_DIR)

                execoutput_file = self.head_explorer.get_file_pathname(\
                                                           execoutput_file_key)
            else:
                execoutput_file = None
                                    
            # @Checkpointing
            if task_untouched:
                if self.meta_testcase_tool.has_checkpointer():
                    self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

                self.head_explorer.remove_file_and_get(matrix_file_key)
                self.head_explorer.remove_file_and_get(execoutput_file_key)
                self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .restart_task()
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # Execute tests
            test_list_file = self.head_explorer.get_file_pathname(\
                                        outdir_struct.TMP_SELECTED_TESTS_LIST)

            meta_testcases = common_fs.loadJSON(test_list_file)

            # Check pass fail
            self.meta_testcase_tool.runtests(meta_testcases=meta_testcases, \
                        stop_on_failure=\
                                self.config.STOP_TESTS_EXECUTION_ON_FAILURE\
                                                                .get_val(), \
                        recalculate_execution_times=True, \
                        fault_test_execution_matrix_file=matrix_file, \
                        fault_test_execution_execoutput_file=execoutput_file, \
                        test_prioritization_module=\
                                        self.meta_testexec_optimization_tool, \
                        finish_destroy_checkpointer=False)
            
            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())
            # Destroy meta test checkpointer
            self.meta_testcase_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

        elif task == checkpoint_tasks.Tasks.CRITERIA_GENERATION_GUIDANCE:
            pass #TODO (Maybe could be used someday. for now, just skip it)
        elif task == checkpoint_tasks.Tasks.CRITERIA_GENERATION:
            # @Checkpointing
            if task_untouched:
                if self.meta_criteria_tool.has_checkpointer():
                    self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

                self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .restart_task()
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            self.meta_criteria_tool.instrument_code(criteria_enabled_list=\
                                    self.config.ENABLED_CRITERIA.get_val(), \
                                    finish_destroy_checkpointer=False)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())
            # Destroy meta test checkpointer
            self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()
        elif task == checkpoint_tasks.Tasks.\
                                CRITERIA_EXECUTION_SELECTION_PRIORITIZATION:
            pass #TODO (Maybe could be used someday. for now, just skip it)
            #TODO: make it as a driver that will implement techniques 
            # (per tool if needed)
            out_file_key = outdir_struct.TMP_SELECTED_CRITERIA_OBJECTIVES_LIST
            out_file = self.head_explorer.get_file_pathname(out_file_key)
                                        
            # @Checkpointing
            if task_untouched:
                if self.meta_criteria_tool.has_checkpointer():
                    self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()
                self.head_explorer.remove_file_and_get(out_file_key)
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            selected_TO = {crit.get_str(): None \
                            for crit in self.config.ENABLED_CRITERIA.get_val()}
            for crit, sel_tech in \
                        self.config.CRITERIA_ELEM_SELECTIONS.get_val().items():
                info_obj = self.meta_criteria_tool\
                                            .get_criterion_info_object(crit)
                if info_obj is None:
                    continue
                all_to = info_obj.get_elements_list()
                
                sel_count = self.config\
                            .MAX_CRITERIA_ELEM_SELECTION_NUM_PERCENT.get_val()
                if type(sel_count) == str:
                    if sel_count.endswith('%'):
                        sel_count = float(sel_count[:-1])
                        ERROR_HANDLER.assert_true(sel_count > 0 \
                                                    and sel_count <= 100, \
                                "invalid selection percentage ({})".format(\
                                                        sel_count), __file__)
                        sel_count = int(math.ceil(\
                                            len(all_to) * sel_count / 100.0))
                    else:
                        ERROR_HANDLER.assert_true(sel_count.isdigit(), \
                            'invalid selection number ({})'.format(sel_count),\
                                                                    __file__)
                        sel_count = int(sel_count)
                else:
                    ERROR_HANDLER.assert_true(type(sel_count) == int, \
                            "invalid selection number ({})".format(sel_count),\
                                                                    __file__)
                ERROR_HANDLER.assert_true(sel_count > 0, \
                                "selection number must be positive", __file__)

                # selection method
                if inspect.isfunction(sel_tech):
                    ERROR_HANDLER.assert_true(\
                                        len(inspect.getargspec().args) == 2, \
                                        "Selection function must take 2 args "
                                        "(testobjectivelist, maxselcount)", \
                                        __file__)
                    
                else:
                    if sel_tech != 'DummyRandom':
                        ERROR_HANDLER.error_exit(\
                            'only random TO selection supported now!', \
                                                                    __file__)
                    # make DummyRandom selection
                    sel_tech = random.sample

                # make selection
                selected_TO[crit.get_str()] = sel_tech(all_to, sel_count)
            
            # write down selection
            common_fs.dumpJSON(selected_TO, out_file)

        elif task == checkpoint_tasks.Tasks.CRITERIA_TESTS_EXECUTION:
            # Make sure that the Matrices dir exists
            self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.RESULTS_MATRICES_DIR)

            matrix_files_keys = {}
            matrix_files = {}
            execoutput_files_keys = {}
            execoutput_files = {}
            for criterion in self.config.ENABLED_CRITERIA.get_val():
                matrix_files_keys[criterion] = \
                                outdir_struct.TMP_CRITERIA_MATRIX[criterion]
                matrix_files[criterion] = \
                                self.head_explorer.get_file_pathname(\
                                                matrix_files_keys[criterion])
                execoutput_files_keys[criterion] = \
                        outdir_struct.TMP_CRITERIA_EXECUTION_OUTPUT[criterion]
                if criterion in self.config.CRITERIA_WITH_OUTPUT_SUMMARY\
                                                                    .get_val():
                    # Make sure that the Exec output dir exists
                    self.head_explorer.get_or_create_and_get_dir(\
                            outdir_struct.RESULTS_TESTEXECUTION_OUTPUTS_DIR)

                    execoutput_files[criterion] = \
                                self.head_explorer.get_file_pathname(\
                                            execoutput_files_keys[criterion])
                else:
                    execoutput_files[criterion] = None

            # @Checkpointing
            if task_untouched:
                if self.meta_criteria_tool.has_checkpointer():
                    self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .destroy_checkpoint()

                for _, matrix_file_key in list(matrix_files_keys.items()):
                    self.head_explorer.remove_file_and_get(matrix_file_key)
                for _, execoutput_file_key in list(\
                                                execoutput_files_keys.items()):
                    self.head_explorer.remove_file_and_get(execoutput_file_key)
                self.meta_criteria_tool.get_checkpoint_state_object()\
                                                        .restart_task()
                self.cp_data.tasks_obj.set_task_executing(task)
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # XXX: Criteria element execution selection loading
            crit_TO_list_by_crit = None
            if self.config.ONLY_EXECUTE_SELECTED_CRITERIA_ELEM.get_val():
                criteria_TO_sel_file = self.head_explorer.get_file_pathname(\
                        outdir_struct.TMP_SELECTED_CRITERIA_OBJECTIVES_LIST)
                raw_crit2to_list = common_fs.loadJSON(criteria_TO_sel_file)
                crit_TO_list_by_crit = {}
                for crit in self.config.ENABLED_CRITERIA.get_val():
                    crit_TO_list_by_crit[crit] = \
                                            raw_crit2to_list[crit.get_str()]

            # Get sequence
            criteria_set_sequence = self.config.CRITERIA_SEQUENCE.get_val()
            #if criteria_set_sequence is None:
            #    criteria_set_sequence = criteria_pkg.CRITERIA_SEQUENCE
            for cs_pos, criteria_set in enumerate(criteria_set_sequence):
                criteria_set &= set(matrix_files)
                if len(criteria_set) == 0:
                    continue

                # ensure right criteria
                used_crit_TO_list_by_crit = copy.deepcopy(crit_TO_list_by_crit)
                if used_crit_TO_list_by_crit is not None:
                    todel = set(used_crit_TO_list_by_crit) - criteria_set
                    for td in todel:
                        del used_crit_TO_list_by_crit[td]

                # Was it already checkpointed w.r.t criteria set seq
                if self.cp_data.criteria_set_is_executed(cs_pos, criteria_set):
                    continue
                
                # If we have a new criteria set id
                self.cp_data.switchto_new_criteria_set(cs_pos, criteria_set)

                # get matrices by criteria
                test_list_file = self.head_explorer.get_file_pathname(\
                                        outdir_struct.TMP_SELECTED_TESTS_LIST)
                meta_testcases = common_fs.loadJSON(test_list_file)
                
                criterion_to_matrix = {\
                                    c: matrix_files[c] for c in criteria_set}

                # TODO: check set based on the criteria for which execout is enabled
                criterion_to_execoutput = {\
                                c: execoutput_files[c] for c in criteria_set}

                # TODO: make recovery mechanism (like atomic) if execution
                #           is interrupted after the matrix is written 
                #           but before checkpoint
                #       The same goes for after the matrix is written to 
                #       'update_matrix_to_cover_when_difference'

                # XXX execute
                self.meta_criteria_tool.runtests_criteria_coverage( \
                            testcases=meta_testcases, \
                            criterion_to_matrix=criterion_to_matrix, \
                            criterion_to_executionoutput=\
                                                    criterion_to_execoutput, \
                            criteria_element_list_by_criteria=\
                                                    used_crit_TO_list_by_crit,
                            cover_criteria_elements_once=self.config.\
                                    COVER_CRITERIA_ELEMENTS_ONCE.get_val(),\
                            prioritization_module_by_criteria=\
                                    self.meta_criteriaexec_optimization_tools,\
                            finish_destroy_checkpointer=True)

                # Update matrix if needed to have output diff or such
                for crit in criteria_set & set(self.config\
                                .CRITERIA_REQUIRING_OUTDIFF_WITH_PROGRAM\
                                                                .get_val()):
                    pf_matrix_file = self.head_explorer.get_file_pathname(\
                                    outdir_struct.TMP_TEST_PASS_FAIL_MATRIX)
                    if not os.path.isfile(pf_matrix_file):
                        #In case already did stats
                        pf_matrix_file = self.head_explorer.get_file_pathname(\
                                    outdir_struct.TEST_PASS_FAIL_MATRIX)
                    if self.config.GET_PASSFAIL_OUTPUT_SUMMARY.get_val():
                        pf_execoutput_file = \
                                        self.head_explorer.get_file_pathname(\
                                outdir_struct.TMP_PROGRAM_TESTEXECUTION_OUTPUT)
                        if not os.path.isfile(pf_execoutput_file):
                            pf_execoutput_file = \
                                        self.head_explorer.get_file_pathname(\
                                outdir_struct.PROGRAM_TESTEXECUTION_OUTPUT)
                    else:
                        pf_execoutput_file = None
                    DriversUtils.update_matrix_to_cover_when_difference(\
                                                criterion_to_matrix[crit], \
                                            criterion_to_execoutput[crit], \
                                            pf_matrix_file, pf_execoutput_file)

                # @Checkpointing
                self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

        elif task == checkpoint_tasks.Tasks.PASS_FAIL_STATS:
            # Make sure that the Matrices dir exists
            self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.RESULTS_STATS_DIR)

            tmp_matrix_file = self.head_explorer.get_file_pathname(\
                                    outdir_struct.TMP_TEST_PASS_FAIL_MATRIX)
            matrix_file = self.head_explorer.get_file_pathname(\
                                    outdir_struct.TEST_PASS_FAIL_MATRIX)
            tmp_execoutput_file = self.head_explorer.get_file_pathname(\
                                outdir_struct.TMP_PROGRAM_TESTEXECUTION_OUTPUT)
            execoutput_file = self.head_explorer.get_file_pathname(\
                                    outdir_struct.PROGRAM_TESTEXECUTION_OUTPUT)
            # Merge TMP pass fail and potentially existing pass fail
            StatsComputer.merge_lmatrix_into_right(\
                                                tmp_matrix_file, matrix_file)
            if self.config.GET_PASSFAIL_OUTPUT_SUMMARY.get_val():
                StatsComputer.merge_lexecoutput_into_right(\
                                        tmp_execoutput_file, execoutput_file)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # Cleanup
            self.head_explorer.remove_file_and_get(\
                                    outdir_struct.TMP_TEST_PASS_FAIL_MATRIX)
            if self.config.GET_PASSFAIL_OUTPUT_SUMMARY.get_val():
                self.head_explorer.remove_file_and_get(\
                                outdir_struct.TMP_PROGRAM_TESTEXECUTION_OUTPUT)

        elif task == checkpoint_tasks.Tasks.CRITERIA_STATS:
            # Make sure that the Matrices dir exists
            self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.RESULTS_STATS_DIR)

            # Merge TMP criteria and potentially existing criteria
            for criterion in self.config.ENABLED_CRITERIA.get_val():
                tmp_matrix_file = self.head_explorer.get_file_pathname(\
                                outdir_struct.TMP_CRITERIA_MATRIX[criterion])
                matrix_file = self.head_explorer.get_file_pathname(\
                                outdir_struct.CRITERIA_MATRIX[criterion])
                tmp_execoutput_file = self.head_explorer.get_file_pathname(\
                        outdir_struct.TMP_CRITERIA_EXECUTION_OUTPUT[criterion])
                execoutput_file = self.head_explorer.get_file_pathname(\
                            outdir_struct.CRITERIA_EXECUTION_OUTPUT[criterion])
                StatsComputer.merge_lmatrix_into_right(tmp_matrix_file, \
                                                                matrix_file)
                if criterion in self.config.CRITERIA_WITH_OUTPUT_SUMMARY\
                                                                    .get_val():
                    StatsComputer.merge_lexecoutput_into_right(\
                                        tmp_execoutput_file, execoutput_file)

            # @Checkpointing
            self.cp_data.tasks_obj.set_task_completed(task)
            self.checkpointer.write_checkpoint(self.cp_data.get_json_obj())

            # Cleanup
            for criterion in self.config.ENABLED_CRITERIA.get_val():
                self.head_explorer.remove_file_and_get(\
                                outdir_struct.TMP_CRITERIA_MATRIX[criterion])

        elif task == checkpoint_tasks.Tasks.AGGREGATED_STATS:
            # Get the test list and criteria element list
            self.meta_testcase_tool.get_testcase_info_file()
            for criterion in self.config.ENABLED_CRITERIA.get_val():
                self.meta_criteria_tool.get_criterion_info_file(criterion)

            # Other results
            other_res = self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.OTHER_COPIED_RESULTS)
            flake_dir = self.meta_testcase_tool.get_flakiness_workdir()
            if os.path.isdir(flake_dir):
                shutil.copytree(flake_dir, os.path.join(other_res, \
                                                os.path.basename(flake_dir)))

            # Compute the final stats (MS, ...)
            StatsComputer.compute_stats(self.config, self.head_explorer, \
                                                            self.checkpointer)
        #-------------------------------------------------------------------

        if not self.cp_data.tasks_obj.task_is_complete(task):
            # @Checkpoint: set task as done
            self.cp_data.tasks_obj.set_task_completed(task)

            # @Checkpoint: write checkpoint
            self.checkpointer.write_checkpoint(\
                                        json_obj=self.cp_data.get_json_obj())
    #~ def _execute_task()

    @classmethod
    def create_repo_manager(cls, config):
        repo_mgr = RepositoryManager(\
                    repository_rootdir=config.REPOSITORY_ROOT_DIR.get_val(),\
                    repo_executables_relpaths=\
                            config.REPO_EXECUTABLE_RELATIVE_PATHS.get_val(),\
                    dev_test_runner_func=\
                            config.CUSTOM_DEV_TEST_RUNNER_FUNCTION.get_val(),\
                    dev_test_program_wrapper=\
                            config.CUSTOM_DEV_TEST_PROGRAM_WRAPPER_CLASS.\
                                                                    get_val(),\
                    test_exec_output_cleaner_func=config\
                                .CUSTOM_TEST_EXECUTION_OUTPUT_CLEANER_FUNCTION\
                                .get_val(), \
                    code_builder_func=config.CODE_BUILDER_FUNCTION.get_val(),\
                    source_files_to_objects=\
                        config.TARGET_SOURCE_INTERMEDIATE_CODE_MAP.get_val(),\
                    dev_tests_list=config.DEVELOPER_TESTS_LIST.get_val(),\
                    )
        return repo_mgr
    #~ def create_repo_manager()

    def _create_meta_test_tool(self, config, head_explorer):
        # create and return the metatest_tool
        meta_test_tool = MetaTestcaseTool(\
                        language=config.PROGRAMMING_LANGUAGE.get_val(),\
                        tests_working_dir=self.head_explorer.get_dir_pathname(\
                                            outdir_struct.TESTSCASES_WORKDIR),\
                        code_builds_factory=self.cb_factory,
                        test_tool_config_list=\
                                    config.TESTCASE_TOOLS_CONFIGS.get_val(),\
                        head_explorer=head_explorer, \
                        hash_outlog=config.HASH_OUTLOG.get_val())
        return meta_test_tool
    #~ def _create_meta_test_tool()

    def _create_meta_criteria_tool(self, config, testcase_tool):
        meta_criteria_tool = MetaCriteriaTool(\
                    language=config.PROGRAMMING_LANGUAGE.get_val(),\
                    meta_test_generation_obj=testcase_tool,\
                    criteria_working_dir=\
                                        self.head_explorer.get_dir_pathname(\
                                            outdir_struct.CRITERIA_WORKDIR),\
                    code_builds_factory=self.cb_factory,
                    tools_config_by_criterion_dict=\
                        config.CRITERIA_TOOLS_CONFIGS_BY_CRITERIA.get_val(),)
        return meta_criteria_tool
    #~ def _create_meta_criteria_tool()

    def _create_meta_testgen_guidance(self, config):
        # TODO
        #ERROR_HANDLER.error_exit("To be implemented", __file__)
        tgg_tool = None
        return tgg_tool
    #~ def _create_meta_testgen_guidance()

    def _create_meta_testexec_optimization(self, config):
        # TODO
        #ERROR_HANDLER.error_exit("To be implemented", __file__)
        teo_tool = None
        return teo_tool
    #~ def _create_meta_testexec_optimization()

    def _create_meta_criteriagen_guidance(self, config):
        # TODO
        #ERROR_HANDLER.error_exit("To be implemented", __file__)
        cgg_tool = None
        return cgg_tool
    #~ def _create_meta_criteriagen_guidance()

    def _create_meta_criteriaexec_optimization(self, config):
        ceo_tools = None
        if len(config.CRITERIA_EXECUTION_OPTIMIZERS.get_val()) > 0:
            ceo_tools = {}
            for crit, opt in \
                        config.CRITERIA_EXECUTION_OPTIMIZERS.get_val().items():
                ceo_tools[crit] = opt.get_optimizer()\
                                            (config, self.head_explorer, crit) 
        return ceo_tools
    #~ def _create_meta_criteriaexec_optimization()

    def _initialize_output_structure(self, cleanstart=False):
        if cleanstart:
            if common_mix.confirm_execution(
                                    "Do you really want to clean the outdir?"):
                self.head_explorer.clean_create_and_get_dir(\
                                            outdir_struct.TOP_OUTPUT_DIR_KEY)
            else:
                ERROR_HANDLER.error_exit("Cancelled Cleanstart!", __file__)
        else:
            self.head_explorer.get_or_create_and_get_dir(\
                                            outdir_struct.TOP_OUTPUT_DIR_KEY)

        for folder in [outdir_struct.CONTROLLER_DATA_DIR, \
                                    outdir_struct.CTRL_CHECKPOINT_DIR, \
                                    outdir_struct.CTRL_LOGS_DIR, \
                                    outdir_struct.EXECUTION_TMP_DIR]:
            self.head_explorer.get_or_create_and_get_dir(folder)
    #~ def _initialize_output_structure()

    def _get_checkpoint_files(self):
        cp_file = self.head_explorer.get_file_pathname(\
                                        outdir_struct.EXECUTION_STATE)
        cp_file_bak = self.head_explorer.get_file_pathname(\
                                        outdir_struct.EXECUTION_STATE_BAKUP)
        return (cp_file, cp_file_bak)
    #~ def _get_checkpoint_files()
#~ class Executor




from __future__ import print_function

import os
import sys
import itertools
import copy
import pandas as pd

import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

ERROR_HANDLER = common_mix.ErrorHandler

DEFAULT_KEY_COLUMN_NAME="MUTERIA_MATRIX_KEY_COL"
class RawExecutionMatrix(object):
    '''
        A 2D matrix representation of the execution of entities by test cases.
        Each entity is represented as a 'key' (The keys are unique) 
        on each row, and, the test cases represented as columns. 
        The behavior of each entity with regard to each test case 
        (affected or not) is represented as active or not. 
        The cell of the matrix with row key E and 
        column test case T represents whether T affects E.

        This class represent a configurable implementation of the Matrix.


        Parameters
        ----------
        filename : file system path string representing the filename where the 
                    underlying pandas dataframe of the matrix is stored
        key_column_name : The string representing the pandas dataframe column 
                        name that will be used to represent the entity ('key')
        non_key_col_list : list containing the list of strings representing
                        the test cases (pandas dataframe column that are not
                        used as key)
        active_cell_default_val : default value of the matrix cell among
                        those considered as active
        inactive_cell_vals : list containing the matrix cell values that are 
                        considered as inactive
        uncertain_cell_default_val : default value of the matrix cell among
                        those considered as uncertain
        is_active_cell_func : function that check whether a cell is active.
                        It takes the cell value as single parameter and returns
                        True if active and False otherwise
        is_inactive_cell_func : function that check whether a cell is inactive.
                        It takes the cell value as single parameter and returns
                        True if active and False otherwise
        is_uncertain_cell_func : function that check whether a cell is N/A.
                        It takes the cell value as simgle parameter and returns
                        True if uncertain (N/A) and False otherwise
        cell_dtype : data type that will be used in underlying pandas dataframe
                        to represent a cell.

        Note
        ---- 
        To ensure consistency in cell type and others, define an extension of 
        this class with fixed values except filename and non_key_col_list
        see for example ExecutionMatrix below.
    '''

    def __init__(self, filename=None, key_column_name=DEFAULT_KEY_COLUMN_NAME,
                    non_key_col_list=None, active_cell_default_val=[1], 
                    inactive_cell_vals=[0], uncertain_cell_default_val=[-1],
                    is_active_cell_func=lambda x: x > 0, 
                    is_inactive_cell_func=lambda x: x == 0, 
                    is_uncertain_cell_func=lambda x: x < 0,
                    cell_dtype=int):
        self.filename = filename
        self.key_column_name = key_column_name
        self.non_key_col_list = non_key_col_list
        self.active_cell_default_val = active_cell_default_val
        self.inactive_cell_vals = inactive_cell_vals
        self.uncertain_cell_default_val = uncertain_cell_default_val
        self.is_active_cell_func = is_active_cell_func
        self.is_inactive_cell_func = is_inactive_cell_func
        self.is_uncertain_cell_func = is_uncertain_cell_func

        # Verify 
        ERROR_HANDLER.assert_true(self.is_active_cell_func(\
                                self.active_cell_default_val[0]), "", __file__)
        ERROR_HANDLER.assert_true(self.is_inactive_cell_func(\
                                    self.inactive_cell_vals[0]), "", __file__) 
        ERROR_HANDLER.assert_true(self.is_uncertain_cell_func(\
                            self.uncertain_cell_default_val[0]), "", __file__)
        # Verify inactive
        for v in self.inactive_cell_vals:
            ERROR_HANDLER.assert_true(not self.is_active_cell_func(v), \
                                                                "", __file__)
        for v in self.inactive_cell_vals:
            ERROR_HANDLER.assert_true(not self.is_uncertain_cell_func(v), \
                                                                "", __file__)

        if self.filename is None or not os.path.isfile(self.filename):
            ERROR_HANDLER.assert_true(self.non_key_col_list is not None, \
                                    "Must specify 'non_key_col_list' when " + \
                                    "filename inexistant. filename is " +
                                    str(self.filename), __file__)
            ordered_cols = [self.key_column_name] + self.non_key_col_list
            self.dataframe = \
                    pd.DataFrame({c:[] for c in ordered_cols})[ordered_cols]
            self.dataframe = self.dataframe.astype(cell_dtype)\
                                           .astype({self.key_column_name: str})
        else:
            self.dataframe = common_fs.loadCSV(self.filename)
            #ERROR_HANDLER.assert_true(len(self.dataframe.columns) >= 2, \
            #        "expect at least 2 columns in dataframe: key, values...",\
            #                                                        __file__)
            ERROR_HANDLER.assert_true(self.key_column_name == \
                                    list(self.dataframe)[0], "key_column"\
                            " name missing or not first column in dataframe",\
                                                                    __file__)
            if self.non_key_col_list is None:
                self.non_key_col_list = list(self.dataframe)[1:]
            else:
                ERROR_HANDLER.assert_true(self.non_key_col_list == \
                                list(self.dataframe)[1:], "non key mismatch",\
                                                                    __file__)

        self.keys_set = set(self.get_keys())

    def get_a_deepcopy(self, new_filename=None, serialize=True):
        """ get a copy of the current matrix. The new filename will be 
            used fo storage.
        :param new_filename: filename to store the copy
        :param serialize: (bool) decide whether to serialize the copy to 
                            file upon creation
        :return: a deep copy of this matrix

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> c_mat = mat.get_a_deepcopy()
        >>> mat.dataframe.equals(c_mat.dataframe)
        True
        >>> mat.add_row_by_key('r', [1, 2, 3])
        >>> mat.dataframe.equals(c_mat.dataframe)
        False
        """
        ret_matrix = copy.deepcopy(self)
        ret_matrix.filename = new_filename
        if serialize:
            ret_matrix.serialize()
        return ret_matrix

    def getActiveCellDefaultVal(self):
        """ Get a value that represent an active cell value
        """
        return self.active_cell_default_val[0]
        
    def getInactiveCellVal(self):
        """ Get a value that represent an inactive cell value
        """
        return self.inactive_cell_vals[0]

    def getUncertainCellDefaultVal(self):
        """ Get a value that represent an uncertain cell value
        """
        return self.uncertain_cell_default_val[0]

    def serialize(self):
        """ Serialize the matrix to its corresponding file if not None
        """
        if self.filename is not None:
            common_fs.dumpCSV(self.dataframe, self.filename)
    #~ def serialize()

    def get_store_filename(self):
        """ Get the name of the storing file
        """
        return self.filename
    #~ def get_store_filename()

    #def raw_add_row(self):

    #def raw_delete_row(self):

    def add_row_by_key(self, key, values, serialize=True):
        """ add a row to the matrix
        :param key: The key to add
        :param values: (list or dict) The values for the given key.
                        Ordered following non key columns ordering if list
                        if dict, it is mapping from column name to value
        :param serialize: (bool) decide whether to serialize the matrix to 
                        file after adding
        :return: nothing

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat._get_key_values_dict(['k']) == {'k': {'a':1,'b':2,'c':3}}
        True
        >>> mat.add_row_by_key('r', {'a':3, 'c':2, 'b':3})
        >>> mat._get_key_values_dict(['r']) == {'r': {'a':3,'b':3,'c':2}}
        True
        >>> len(mat.get_keys())
        2
        """
        ERROR_HANDLER.assert_true(key not in self.keys_set, \
                            "adding an existing key: '"+str(key)+\
                            "', to matrix: " + str(self.filename), __file__)
        self.keys_set.add(key)
        if type(values) in (list, tuple):
            self.dataframe.loc[len(self.dataframe)] = [key] + values
        elif type(values) == dict:
            ERROR_HANDLER.assert_true(self.key_column_name not in values, \
                                        "key column name in values", __file__)
            tmpdict = {self.key_column_name: key}
            tmpdict.update(values)
            self.dataframe = self.dataframe.append(tmpdict, ignore_index=True)
        else:
            ERROR_HANDLER.error_exit("Invald input: 'values'", __file__)

        if serialize:
            self.serialize()

    def delete_rows_by_key(self, key_list, serialize=True):
        """ delete the rows for the given keys
        :param key_list: collection of key whose rows to delete
        :param serialize: (bool) decide whether to serialize the matrix
                         to file after deletion
        :return: nothing

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.add_row_by_key('r', [3, 2, 3])
        >>> mat.add_row_by_key('w', [1, 0, 3])
        >>> mat.delete_rows_by_key(['k','w'])
        >>> mat_keys = mat.get_keys()
        >>> len(mat_keys)
        1
        >>> mat_keys[0] == 'r'
        True
        """
        self.dataframe = \
            self.dataframe.loc[~self.dataframe[self.key_column_name]\
                                                        .isin(set(key_list))]
        self.dataframe = self.dataframe.reset_index(drop=True)

        if serialize:
            self.serialize()

    def to_pandas_df(self):
        """ return the matrix as a pandas dataframe 
            (just return a copy of the underlying pandas dataframe)
        :return: a deep copy of the underlying pandas dataframe 
                (modifying the dataframe do ot affect the matrix)

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> df2 = mat.to_pandas_df()
        >>> mat.add_row_by_key('r', [3, 2, 3])
        >>> len(df2)
        1
        >>> len(mat.get_keys())
        2
        """
        return self.dataframe.copy(deep=True)

    def get_key_colname(self):
        """ get the name of the column representing the keys
        """
        return self.key_column_name

    def get_nonkey_colname_list(self):
        """ get list of non key columns
        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.get_nonkey_colname_list() == nc
        True
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.get_nonkey_colname_list() == nc
        True
        """
        return self.non_key_col_list

    def get_keys(self):
        """ get a pandas serie of the keys (example mutant ids)
        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> type(mat.get_keys()) == pd.core.series.Series
        True
        >>> len(mat.get_keys())
        0
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> list(mat.get_keys())
        ['k']
        """
        return self.dataframe[self.key_column_name]

    def is_empty(self):
        """ Check that the matrix have no row (all columns have no row)
        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.is_empty()
        True
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.is_empty()
        False
        """
        return len(self.get_keys()) == 0

    def extract_by_rowkey(self, row_key_list, out_filename=None):
        """ get the sub-matrix with the rows keys corresponding to the
            values in the passed list.
            Note that the order of keys in row_key_list is not guaranted
        :param row_key_list: collection of row keys to extract. 
                            Must not be empty and every key must exist
        :param out_filename: Optional filename to use as storage for the 
                            extracted matrix. Default is None (no storage)
        :return: matrix which is sub matrix of this

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.add_row_by_key('r', [3, 2, 3])
        >>> mat.add_row_by_key('w', [1, 0, 3])
        >>> mat2 = mat.extract_by_rowkey(['r', 'k'])
        >>> list(mat2.get_keys()) == ['k', 'r']
        True
        >>> mat2.get_nonkey_colname_list() == nc
        True
        """
        ERROR_HANDLER.assert_true(len(row_key_list) > 0, \
                                    "key list should not be empty", __file__)
        ERROR_HANDLER.assert_true(len(set(row_key_list) - self.keys_set) == 0,\
                                    "invalid row key passed to extract row", \
                                                                    __file__)

        ret = self.get_a_deepcopy(new_filename=out_filename, serialize=True)
        ret.dataframe = ret.dataframe.loc[ret.dataframe[ret.key_column_name]\
                                                        .isin(row_key_list)]
        ret.keys_set = set(ret.get_keys())
        return ret

    def extract_by_column(self, non_key_col_list, out_filename=None):
        """ get the sub-matrix with the columns corresponding to the
            values in the passed list.
            Note that the order is maintained if a list is passed
        :param non_key_col_list: collection of columns to extract. 
                            Must not be empty and every column must exist
        :param out_filename: Optional filename to use as storage for the 
                            extracted matrix. Default is None (no storage)
        :return: matrix which is sub matrix of this

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.add_row_by_key('r', [3, 2, 3])
        >>> mat.add_row_by_key('w', [1, 0, 3])
        >>> mat2 = mat.extract_by_column(['a', 'c'])
        >>> list(mat2.get_keys()) == ['k', 'r', 'w']
        True
        >>> mat2.get_nonkey_colname_list() == ['a', 'c']
        True
        """
        ERROR_HANDLER.assert_true(len(non_key_col_list) > 0, \
                                    "col list should not be empty", __file__)
        ERROR_HANDLER.assert_true(len(set(non_key_col_list) - \
                                        set(self.non_key_col_list)) == 0, \
                                    "invalid column passed to extract col", \
                                                                    __file__)
        non_key_col_list = list(non_key_col_list)
        ret = self.get_a_deepcopy(new_filename=out_filename, serialize=True)
        ret.dataframe = ret.dataframe[[ret.key_column_name]+non_key_col_list]
        ret.non_key_col_list = non_key_col_list
        return ret

    def query_active_columns_of_rows(self, row_key_list=None):
        ''' return a dict in the form row2cols
        :param row_key_list: list of rows to query for
        :return: a dict representing a mapping between the passed rows
                and the list of columns active for those rows

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_active_columns_of_rows() == {'k': ['c']}
        True
        '''
        if row_key_list is None:
            row_key_list = self.get_keys()

        result = {}
        if len(row_key_list) > 0:
            small_df = self.extract_by_rowkey(row_key_list).dataframe
            for _, row in small_df.iterrows():
                result[row[self.key_column_name]] = \
                                        [x for x in self.non_key_col_list \
                                        if self.is_active_cell_func(row[x])]

        return result
    #~ def query_active_columns_of_rows()

    def query_active_rows_of_columns(self, non_key_col_list=None):
        ''' return a dict in the form col2rows
        :param non_key_col_list: list of columns to query for
        :return: a dict representing a mapping between the passed columns
                and the list of rows active for those columns

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_active_rows_of_columns() == {'a':[], 'b':[], 'c':['k']}
        True
        '''
        if non_key_col_list is None:
            non_key_col_list = self.non_key_col_list

        result = {}
        if len(non_key_col_list) > 0:
            small_df = self.extract_by_column(non_key_col_list).dataframe
            for col in non_key_col_list:
                result[col] = list(small_df.loc[\
                                small_df[col].apply(self.is_active_cell_func)]\
                                                        [self.key_column_name])

        return result
    #~ def query_active_rows_of_columns()

    def query_inactive_columns_of_rows(self, row_key_list=None):
        ''' return a dict in the form row2cols
        :param row_key_list: list of rows to query for
        :return: a dict representing a mapping between the passed rows
                and the list of columns inactive for those rows

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_inactive_columns_of_rows() == {'k': ['a']}
        True
        '''
        if row_key_list is None:
            row_key_list = self.get_keys()

        result = {}
        if len(row_key_list) > 0:
            small_df = self.extract_by_rowkey(row_key_list).dataframe
            for _, row in small_df.iterrows():
                result[row[self.key_column_name]] = \
                                        [x for x in self.non_key_col_list \
                                        if self.is_inactive_cell_func(row[x])]

        return result
    #~ def query_inactive_columns_of_rows()

    def query_inactive_rows_of_columns(self, non_key_col_list=None):
        ''' return a dict in the form col2rows
        :param non_key_col_list: list of columns to query for
        :return: a dict representing a mapping between the passed columns
                and the list of rows inactive for those columns

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_inactive_rows_of_columns() == {'a':['k'], 'b':[], 'c':[]}
        True
        '''
        if non_key_col_list is None:
            non_key_col_list = self.non_key_col_list

        result = {}
        if len(non_key_col_list) > 0:
            small_df = self.extract_by_column(non_key_col_list).dataframe
            for col in non_key_col_list:
                result[col] = list(small_df.loc[\
                            small_df[col].apply(self.is_inactive_cell_func)]\
                                                        [self.key_column_name])

        return result
    #~ def query_inactive_rows_of_columns()

    def query_uncertain_columns_of_rows(self, row_key_list=None):
        ''' return a dict in the form row2cols
        :param row_key_list: list of rows to query for
        :return: a dict representing a mapping between the passed rows
                and the list of columns uncertain for those rows

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_uncertain_columns_of_rows() == {'k': ['b']}
        True
        '''
        if row_key_list is None:
            row_key_list = self.get_keys()

        result = {}
        if len(row_key_list) > 0:
            small_df = self.extract_by_rowkey(row_key_list).dataframe
            for _, row in small_df.iterrows():
                result[row[self.key_column_name]] = \
                                        [x for x in self.non_key_col_list \
                                        if self.is_uncertain_cell_func(row[x])]

        return result
    #~ def query_uncertain_columns_of_rows()

    def query_uncertain_rows_of_columns(self, non_key_col_list=None):
        ''' return a dict in the form col2rows
        :param non_key_col_list: list of columns to query for
        :return: a dict representing a mapping between the passed columns
                and the list of rows inactive for those columns

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> act = mat.getActiveCellDefaultVal()
        >>> inact = mat.getInactiveCellVal()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> mat.add_row_by_key('k', [inact, uncert, act])
        >>> mat.query_uncertain_rows_of_columns() == {'a':[],'b':['k'],'c':[]}
        True
        '''
        if non_key_col_list is None:
            non_key_col_list = self.non_key_col_list

        result = {}
        if len(non_key_col_list) > 0:
            small_df = self.extract_by_column(non_key_col_list).dataframe
            for col in non_key_col_list:
                result[col] = list(small_df.loc[\
                            small_df[col].apply(self.is_uncertain_cell_func)]\
                                                        [self.key_column_name])

        return result
    #~ def query_uncertain_rows_of_columns()

    def _get_key_values_dict(self, keys=None):
        """ compute a dict object with key each element of keys and value
            the corresponding dict representation of the row(without the key)
        :param keys: list of keys of interest to get values. 
                        If None, all keys are considered
        :return: dict representing a mapping between the row keys and their 
                    row values. 
                    each row value is a dict of column to cell value

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat._get_key_values_dict(['k']) == {'k': {'a':1,'b':2,'c':3}}
        True
        """
        if keys is None:
            keys = self.get_keys()
        if len(keys) == 0:
            k_v_dict = {}
        else:
            tmp_df = self.extract_by_rowkey(keys).dataframe
            tmp_df.set_index(self.get_key_colname(), inplace=True)
            k_v_dict = tmp_df.to_dict('index')
        return k_v_dict

    def update_cells(self, key, values):
        """ Update the values for the key with the values 
        :param key: key whose values to update
        :param values: dict representing the new cell values by column name
        :return: nothing

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.add_row_by_key('r', [1, 2, 3])
        >>> mat.update_cells('k', {'a':0, 'c':4})
        >>> mat._get_key_values_dict(['k']) == {'k': {'a':0, 'b':2, 'c':4}}
        True
        """
        # locate key index
        key_pos = self.dataframe\
                        .index[self.dataframe[self.get_key_colname()] == key]\
                        .tolist()
        ERROR_HANDLER.assert_true(len(key_pos) == 1)
        key_pos = key_pos[0]
        for col, cval in list(values.items()):
            self.dataframe.at[key_pos, col] = cval

    def update_with_other_matrix(self, other_matrix, \
                                override_existing=False, allow_missing=False, \
                                ask_confirmation_with_exist_missing=False, \
                                serialize=False):
        """ Update this matrix using the other matrix
        :param other_matrix: The matrix to use to update this matrix
        :param override_existing: (bool) decide whether existing cell's 
                        value (row, column) of this matrix should be 
                        overrided by the update.
                        Note that if this is disabled, no cell (row, col)
                        of this matrix must exist in other_matrix, or this
                        will fail.
        :param allow_missing: (bool) decide whether missing cell are allowed
                        when merging (cells that only belong to one matrix
                        while thre are other cells on same row or same 
                        column that belong to the two matrices).
                        Note that if this is disable. there should not be
                        any such cell during merging or this will fail.
        :param serialize: (bool) decides whether to serialize the updated
                        matrix (this matrix), after update, into its file.
        :return: nothing

        Example:
        >>> nc = ['a', 'b', 'c']
        >>> mat = ExecutionMatrix(non_key_col_list=nc)
        >>> mat.add_row_by_key('k', [1, 2, 3])
        >>> mat.add_row_by_key('r', [4, 0, 1])
        >>> nc_o = ['b', 'e']
        >>> mat_other = ExecutionMatrix(non_key_col_list=nc_o)
        >>> mat_other.add_row_by_key('h', [200, 100])
        >>> mat_other.add_row_by_key('r', [10, 11])
        >>> mat.update_with_other_matrix(mat_other, override_existing=True, \
                                                            allow_missing=True)
        >>> list(mat.get_keys())
        ['k', 'r', 'h']
        >>> list(mat.get_nonkey_colname_list())
        ['a', 'b', 'c', 'e']
        >>> k_v_d = mat._get_key_values_dict()
        >>> uncert = mat.getUncertainCellDefaultVal()
        >>> k_v_d['k'] == {'a': 1, 'b': 2, 'c': 3, 'e': uncert}
        True
        >>> k_v_d['r'] == {'a': 4, 'b': 10, 'c': 1, 'e': 11}
        True
        >>> k_v_d['h'] == {'a': uncert, 'b': 200, 'c': uncert, 'e': 100}
        True
        """
        #other_matrix_df = other_matrix.to_pandas_df()
        # Check values overlap
        row_existing = set(self.get_keys()) & set(other_matrix.get_keys())
        col_existing = set(self.get_nonkey_colname_list()) & \
                            set(other_matrix.get_nonkey_colname_list())
        if len(row_existing) > 0 and len(col_existing) > 0:
            if ask_confirmation_with_exist_missing and override_existing:
                override_existing = common_mix.confirm_execution(\
                                            "Some cells are existing, "
                                            "do you confirm their override?")
                ERROR_HANDLER.assert_true(override_existing, \
                            "Override_existing not set but there is overlap", \
                                                                    __file__)            

        # Check missing
        if len(row_existing) < len(other_matrix.get_keys()):
            #- Some rows will be added
            if len(col_existing) != len(self.get_nonkey_colname_list()) or \
                            len(col_existing) != \
                                len(other_matrix.get_nonkey_colname_list()):
                if ask_confirmation_with_exist_missing and allow_missing:
                    allow_missing = common_mix.confirm_execution(\
                                            "Some values are missing, "
                                            "do you confirm their presence?")
                ERROR_HANDLER.assert_true(allow_missing, \
                                "allow_missing disable but there are missing",\
                                                                    __file__)

        # Actual update
        ## 1. Create columns that are not in others
        col_to_add = set(other_matrix.get_nonkey_colname_list()) - \
                                            set(self.get_nonkey_colname_list())
        for col in col_to_add:
            self.non_key_col_list.append(col)
            self.dataframe[col] = [self.getUncertainCellDefaultVal()] * \
                                                        len(self.get_keys())
        
        ## 2. Update or insert rows
        extra_cols = set(self.get_nonkey_colname_list()) - \
                                    set(other_matrix.get_nonkey_colname_list())
        missing_extracol_vals = {e_c: self.getUncertainCellDefaultVal() \
                                                        for e_c in extra_cols}
        ### Insert
        new_rows = set(other_matrix.get_keys()) - set(self.get_keys())
        k_v_dict = other_matrix._get_key_values_dict(new_rows)
        for key, values in list(k_v_dict.items()):
            values.update(missing_extracol_vals)
            self.add_row_by_key(key, values, serialize=False)
        ### Update
        k_v_dict = other_matrix._get_key_values_dict(row_existing)
        for key, values in list(k_v_dict.items()):
            #for col in values:
            self.update_cells(key, values)

        if serialize:
            self.serialize()
    #~ def update_with_other_matrix()
#~ class RawExecutionMatrix

class ExecutionMatrix(RawExecutionMatrix):
    '''
        This class is the default extension of the class RawExecutionMatrix. 
    '''
    def __init__(self, filename=None, non_key_col_list=None):
        RawExecutionMatrix.__init__(self, filename=filename, \
                                            non_key_col_list=non_key_col_list)
    #~ def __init__()
#~ class ExecutionMatrix


class OutputLogData(object):
    #OBJECTIVE_ID = "OBJECTIVE_ID"
    #TEST_ID = "TEST_ID"
    OUTLOG_LEN = "OUTLOG_LEN"       # int
    OUTLOG_HASH = "OUTLOG_HASH"     # str
    RETURN_CODE = "RETURN_CODE"     # int
    TIMEDOUT = "TIMEDOUT"           # bool
    #ordered_cols = [OBJECTIVE_ID, TEST_ID, OUTLOG_LEN, OUTLOG_HASH, \
    #                                                            RETURN_CODE]
    Dat_Keys = {OUTLOG_LEN, OUTLOG_HASH, RETURN_CODE, TIMEDOUT}
    
    UNCERTAIN_TEST_OUTLOGDATA = {
                OUTLOG_LEN: common_mix.GlobalConstants.COMMAND_UNCERTAIN,
                OUTLOG_HASH: common_mix.GlobalConstants.COMMAND_UNCERTAIN,
                RETURN_CODE: common_mix.GlobalConstants.COMMAND_UNCERTAIN,
                TIMEDOUT: common_mix.GlobalConstants.COMMAND_UNCERTAIN,
    }

    @classmethod
    def outlogdata_equiv(cls, outlogdata1, outlogdata2):
        if outlogdata1 == cls.UNCERTAIN_TEST_OUTLOGDATA or \
                                outlogdata2 == cls.UNCERTAIN_TEST_OUTLOGDATA:
            return None
        if outlogdata1 != outlogdata2 and not(\
                    outlogdata1[cls.TIMEDOUT] and outlogdata2[cls.TIMEDOUT]):
            return False
        return True
    #~ def outlogdata_diff()

    def __init__(self, filename=None):
        self.filename = filename
        if self.filename is None or not os.path.isfile(self.filename):
            self.data = {}
        else:
            self.data = common_fs.loadJSON(self.filename)
    #~ def __init__()

    def is_empty(self):
        return len(self.data) == 0
    #~ def is_empty()

    def get_zip_objective_and_data(self):
        return self.data.items()
    #~ def get_zip_objective_and_data()

    def add_data (self, data_dict, check_all=True, override_existing=False, 
                                                            serialize=False):
        if check_all:
            ERROR_HANDLER.assert_true(\
                            type(data_dict) == dict and len(data_dict) > 0, \
                                        "expecting a non empty dict", __file__)
            for o, o_obj in data_dict.items():
                if len(o_obj) == 0: 
                    continue
                ERROR_HANDLER.assert_true(\
                            type(o_obj) == dict, \
                            "expecting dict for value data: objective is "+o, \
                                                                    __file__)
                for t, t_obj in o_obj.items():
                    ERROR_HANDLER.assert_true(set(t_obj) == self.Dat_Keys , \
                                "Invalid data for o "+o+' and t '+t, __file__)

        intersect_objective = set(self.data) & set(data_dict)
        onlynew_objective = set(data_dict) - intersect_objective
        if not override_existing:
            for objective in intersect_objective:
                ERROR_HANDLER.assert_true(len(set(self.data[objective]) & \
                                            set(data_dict[objective])) == 0, \
                            "Override_existing not set but there is overlap", \
                                                                    __file__)            
        for objective in intersect_objective:
            self.data[objective].update(data_dict[objective])
        for objective in onlynew_objective:
            self.data[objective] = copy.deepcopy(data_dict[objective])
        if serialize:
            self.serialize()
    #~ def add_data ()

    def update_with_other(self, other_execoutput, \
                                    override_existing=False, serialize=False):
        self.add_data(other_execoutput.data, check_all=False, \
                                        override_existing=override_existing, \
                                        serialize=serialize)
    #~ def update_with_other_matrix()

    def serialize(self):
        """ Serialize the matrix to its corresponding file if not None
        """
        if self.filename is not None:
            common_fs.dumpJSON(self.data, self.filename, pretty=True)
    #~ def serialize()

    def get_store_filename(self):
        """ Get the name of the storing file
        """
        return self.filename
    #~ def get_store_filename()
#~ class OutputLogData



import muteria.common.mix as common_mix
import muteria.common.fs as common_fs

ERROR_HANDLER = common_mix.ErrorHandler

class TestToolType(common_mix.EnumAutoName):
    USE_ONLY_CODE = "StaticTestcaseTool"
    USE_CODE_AND_TESTS = "DynamicTestcaseTool"

    def get_tool_type_classname(self):
        return self.get_field_value()
    #~ def get_tool_type_classname():
#~ class TestToolType

TEST_TOOL_TYPES_SCHEDULING = [
    (TestToolType.USE_ONLY_CODE,), 
    (TestToolType.USE_CODE_AND_TESTS,),
]



from muteria.drivers.criteria.tools_by_languages.c.gcov.gcov \
                                                        import CriteriaToolGCov 

#from .gcov import CriteriaToolGCov

StaticCriteriaTool = CriteriaToolGCov




from __future__ import print_function

import itertools

def getCoupledMutants (mutants_to_killingtests, failingtests, istwoways=False):
    coupled = []
    failingtests = set(failingtests)
    for m in mutants_to_killingtests:
        if len(set(mutants_to_killingtests[m]) - failingtests) == 0:
            if istwoways and len(failingtests) != len(mutants_to_killingtests[m]):
                continue
            coupled.append(m)
    return coupled
    
def getSubsumingMutants (mutants_to_killingtests, clustered=True):
    '''
        :param mutants_to_killingtests: dict having as key the mutant ID 
                                and as value the list of test killing it
        :param clustered: decides whether the subsuming mutants are returned
                        as a list of mutants or list of mutants clusters
        :returns: a tuple of (1) list of equivalent mutants
                    (2) the list of tuple of subsuming mutant (Each tuple
                    contain the mutants that are subsuming each others) or
                    list of all subsuming mutants.
    '''

    equivalent_mutants = []
    subsuming_mutants_clusters = []

    # if empty, do nothing
    if len(mutants_to_killingtests) == 0:
        return equivalent_mutants, subsuming_mutants_clusters
    
    # make sure that the tests killing are sets
    if type(mutants_to_killingtests[mutants_to_killingtests.keys()[0]]) != set:
        mutants_to_killingtests = {m: set(mutants_to_killingtests[m]) \
                                            for m in mutants_to_killingtests}

    # Create cluster list of equi-subsumptions, and equivalent mutants list
    cluster_list = []
    for mutant_id in mutants_to_killingtests:
        if len(mutants_to_killingtests[mutant_id]) == 0:
            equivalent_mutants.append(mutant_id)
        else:
            # can it fit in existing cluster:
            found = False
            for cluster in cluster_list:
                if mutants_to_killingtests[cluster[0]] == \
                                            mutants_to_killingtests[mutant_id]:
                    cluster.append(mutant_id)
                    found = True
                    break

            if not found:
                # Create its own cluster
                cluster_list.append([mutant_id])

    # sort the clusters but increating number of killing tests
    cluster_list.sort(key=lambda x: len(mutants_to_killingtests[x[0]]))

    # for each cluster in order, check if subsumed, else, add
    for cluster in cluster_list:
        subsumed = False
        for subsuming_cluster in subsuming_mutants_clusters:
            # Check subsumption
            if len(mutants_to_killingtests[subsuming_cluster[0]] - \
                    mutants_to_killingtests[cluster[0]]) == 0:
                # cluster is subsumed by subsuming_cluster, do nothing
                subsumed = True
                break

        if not subsumed:
            subsuming_mutants_clusters.append(tuple(cluster))

    if not clustered:
        tmp_list = []
        for scluster in subsuming_mutants_clusters:
            tmp_list += scluster
        subsuming_mutants_clusters = tmp_list

    return equivalent_mutants, subsuming_mutants_clusters


def getCommonSetsSizes_venn (setsElemsDict, setsize_from=None, 
                                setsize_to=None, name_delim='&', 
                                not_common=None):
    '''
        TODO: 
    '''
    if not_common is not None:
        assert type(not_common) == dict and len(not_common) == 0

    res_set = {} 
    if setsize_from is None:
        setsize_from = 1
    if setsize_to is None:
        setsize_to = len(setsElemsDict)
    ordered_keys = list(setsElemsDict)
    for setsize in range(setsize_from, setsize_to+1):
        for set_pos in itertools.combinations(range(len(ordered_keys)), setsize):
            name_key = name_delim.join([ordered_keys[i] for i in set_pos])
            assert name_key not in res_set
            res_set[name_key] = None 
            #print (set_pos, len(set_pos))
            for i in set_pos:
                if res_set[name_key] is None:
                    res_set[name_key] = set(setsElemsDict[ordered_keys[i]])
                else:
                    res_set[name_key] &= setsElemsDict[ordered_keys[i]]

            if not_common is not None:
                assert name_key not in not_common
                not_common[name_key] = {}
                for i in set_pos:
                    name_i = ordered_keys[i]
                    extra_tmp = list(set(setsElemsDict[name_i]) - res_set[name_key])
                    if len(extra_tmp) > 0:
                        not_common[name_key][name_i] = extra_tmp

    #print (res_set)
    res_num = {} 
    for v in res_set:
        res_num[v] = len(res_set[v])
    
    return res_num, res_set
#~ def getCommonSetsSizes_venn()
